import sys
import os
import logging
import traceback
from PySide6.QtCore import Qt, QTranslator
from PySide6.QtWidgets import QApplication, QMessageBox
from jal.widgets.main_window import MainWindow
from jal.db.db import JalDB, JalDBError
from jal.db.settings import JalSettings

#-----------------------------------------------------------------------------------------------------------------------
def exception_logger(exctype, value, tb):
    info = traceback.format_exception(exctype, value, tb)
    logging.fatal(f"EXCEPTION: {info}")
    sys.__excepthook__(exctype, value, tb)


#-----------------------------------------------------------------------------------------------------------------------
def main():
    sys.excepthook = exception_logger
    os.environ['QT_MAC_WANTS_LAYER'] = '1'    # Workaround for https://bugreports.qt.io/browse/QTBUG-87014
    app = QApplication([])

    error = JalDB().init_db()

    translator = QTranslator(app)
    translator.load(JalSettings.path(JalDB.PATH_LANG_FILE))
    app.installTranslator(translator)

    if error.code == JalDBError.OutdatedDbSchema:
        error = JalDB().update_db_schema()

    if error.code != JalDBError.NoError:
        window = QMessageBox()
        window.setAttribute(Qt.WA_DeleteOnClose)
        window.setWindowTitle("JAL: Start-up aborted")
        window.setIcon(QMessageBox.Critical)
        window.setText(error.message)
        window.setInformativeText(error.details)
    else:
        window = MainWindow(translator)
    window.show()

    app.exec()
    app.removeTranslator(translator)


#-----------------------------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    main()



from PySide6.QtCore import QAbstractItemModel
from PySide6.QtWidgets import QTreeView
from PySide6.QtGui import QResizeEvent
from .table_footer import FooterView

# ----------------------------------------------------------------------------------------------------------------------
# File implements TableViewWithFooter class that is a descendant of QTableView class with footer.
# Underlying model should support footerData(section, role) method in order to provide data for the footer.
# The footer is implemented as FooterView class that is derived from QHeaderView.
# ----------------------------------------------------------------------------------------------------------------------
class TreeViewWithFooter(QTreeView):
    def __init__(self, parent_view):
        self._parent_view = parent_view
        super().__init__(parent_view)
        self._footer = FooterView(self, self.header())

    def footer(self) -> FooterView:
        return self._footer

    # Create a bottom margin for footer placement (mirror of a top header margin)
    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        margins = self.viewportMargins()
        self.setViewportMargins(margins.left(), margins.top(), margins.right(), margins.top())
        self._footer.on_header_geometry()

    def setModel(self, model: QAbstractItemModel) -> None:
        super().setModel(model)
        self._footer.setModel(model)

    def setColumnHidden(self, idx: int, hidden: bool) -> None:
        super().setColumnHidden(idx, hidden)
        self._footer.setSectionHidden(idx, hidden)



import logging
from decimal import Decimal
from jal.constants import BookAccount, DepositActions
from jal.db.db import JalDB
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.operations import LedgerTransaction


class JalDeposit(JalDB):
    def __init__(self, id: int = 0):
        super().__init__()
        self._id = id
        self._data = self._read("SELECT account_id, note FROM term_deposits WHERE id=:deposit_id",
                                [(":deposit_id", self._id)], named=True)
        self._account_id = 0 if self._data is None else self._data['account_id']
        self._account = JalAccount(self._account_id)
        self._currency = JalAsset(self._account.currency())
        self._note = '' if self._data is None else self._data['note']
        actions_query = self._exec("SELECT timestamp, action_type, amount FROM deposit_actions "
                                   "WHERE deposit_id=:deposit_id", [(":deposit_id", self._id)])
        self._actions = []
        while actions_query.next():
            self._actions.append(self._read_record(actions_query, named=True, cast=[int, int, Decimal]))

    @classmethod
    # Returns a list of deposits that are opened before and not closed at given timestamp
    def get_term_deposits(cls, timestamp: int) -> list:
        deposits = []
        query = cls._exec(
            "SELECT o.deposit_id FROM deposit_actions o "
            "LEFT JOIN deposit_actions c ON o.action_type=:opening AND c.action_type=:closing and o.deposit_id=c.deposit_id "
            "WHERE o.timestamp<=:timestamp AND c.timestamp>=:timestamp",
            [(":timestamp", timestamp), (":opening", DepositActions.Opening), (":closing", DepositActions.Closing)])
        while query.next():
            deposits.append(JalDeposit(super(JalDeposit, JalDeposit)._read_record(query, cast=[int])))
        return deposits

    # returns name of the deposit as it was given in notes
    def name(self) -> str:
        return self._note

    # Returns currency of the deposit
    def currency(self) -> JalAsset:
        return self._currency

    # Return accumulated money for the deposit at given timestamp
    def balance(self, timestamp: int) -> Decimal:
        balance = Decimal('0')
        query = self._exec("SELECT amount FROM ledger "
                           "WHERE book_account=:book AND op_type=:type AND operation_id=:id AND timestamp<=:timestamp",
                           [(":book", BookAccount.Savings), (":type", LedgerTransaction.TermDeposit),
                            (":id", self._id), (":timestamp", timestamp)])
        while query.next():
            balance += self._read_record(query, cast=[Decimal])
        return balance

    # Return a timestamp of deposit opening
    def start_date(self) -> int:
        opening = [x for x in self._actions if x['action_type']==DepositActions.Opening]
        assert len(opening) == 1
        return opening[0]['timestamp']

    # Return a timestamp of deposit closure
    def end_date(self) -> int:
        opening = [x for x in self._actions if x['action_type'] == DepositActions.Closing]
        assert len(opening) == 1
        return opening[0]['timestamp']

    def open_amount(self) -> Decimal:
        opening = [x for x in self._actions if x['action_type'] == DepositActions.Opening]
        assert len(opening) == 1
        return opening[0]['amount']

    def accrued_interest(self, timestamp) -> Decimal:
        amount = Decimal('0')
        interests = [x for x in self._actions if x['timestamp'] <= timestamp]
        for interest in interests:
            if interest['action_type'] == DepositActions.InterestAccrued:
                amount += interest['amount']
            elif interest['action_type'] == DepositActions.TaxWithheld:
                amount += interest['amount']
            else:
                continue
        return amount

    def close_amount(self) -> Decimal:
        amount = self.open_amount()
        for action in self._actions:
            if action['action_type'] in [DepositActions.TopUp, DepositActions.InterestAccrued]:
                amount += action['amount']
            elif action['action_type'] in [DepositActions.PartialWithdrawal, DepositActions.TaxWithheld]:
                amount -= action['amount']
            elif action['action_type'] in [DepositActions.Opening, DepositActions.Closing, DepositActions.Renewal]:
                continue
            else:
                logging.warning(self.tr("Unexpected deposit action: ") + action['action_type'])
        return amount



from PySide6.QtCore import Qt, QAbstractItemModel, QRect
from PySide6.QtWidgets import QTreeView, QHeaderView, QStyle, QStyleOptionHeaderV2
from PySide6.QtGui import QPainter, QIcon, QFontMetrics

#-----------------------------------------------------------------------------------------------------------------------
# Class that implements Footer for custom TreeViewWithFooter and TableViewWithFooter widgets
# Model should support footerData() method that will provide required data to display in the footer.
class FooterView(QHeaderView):
    def __init__(self, parent: QTreeView, table_header: QHeaderView):
        super().__init__(Qt.Horizontal, parent)
        self._span = {}
        self._sizes = {}
        self._parent = parent
        self._model = None
        self._linked_header = table_header
        self.setSectionResizeMode(QHeaderView.Fixed)
        self._linked_header.geometriesChanged.connect(self.on_header_geometry)
        self._linked_header.sectionResized.connect(self.on_header_resize)
        self._linked_header.sectionMoved.connect(self.on_header_move)

    def setModel(self, model: QAbstractItemModel) -> None:
        self._model = model
        self._model.dataChanged.connect(self.on_model_update)
        super().setModel(model)

    def paintSection(self, painter: QPainter, rect: QRect, idx: int) -> None:
        painter.save()
        opt = QStyleOptionHeaderV2()
        self.initStyleOption(opt)
        self.initStyleOptionForIndex(opt, idx)
        opt.rect = rect
        text = self._model.footerData(idx, role=Qt.DisplayRole)
        opt.text = '' if text is None else text
        font = self._model.footerData(idx, role=Qt.FontRole)
        if font is not None:
            opt.fontMetrics = QFontMetrics(font)
            painter.setFont(font)
        icon = self._model.footerData(idx, role=Qt.DecorationRole)
        opt.iconAlignment = Qt.AlignVCenter
        opt.icon = QIcon() if icon is None else icon
        alignment = self._model.footerData(idx, role=Qt.TextAlignmentRole)
        opt.textAlignment = Qt.AlignCenter | Qt.AlignVCenter if alignment is None else alignment
        self.style().drawControl(QStyle.CE_Header, opt, painter, self)
        painter.restore()

    def on_header_geometry(self):
        cr = self._parent.contentsRect()
        hs = self._linked_header.geometry()
        self.setGeometry(cr.left(), cr.top() + cr.height() - hs.height() + 1, hs.width(), hs.height())

    def on_header_resize(self, section: int, _old_size: int, new_size: int) -> None:
        if section not in self._span:   # Simply set size, if section doesn't span to multiple columns
            self.resizeSection(section, new_size)
            return
        if self._span[section] != section:   # Hide, if section gives its place to another section that is spanned over it
            self.setSectionHidden(section, True)
        size = 0   # Calculate a size of section that is spanned over other columns and resize it
        for i in [k for k, v in self._span.items() if v == self._span[section]]:
            size += new_size if i == section else self._linked_header.sectionSize(i)
        self.resizeSection(self._span[section], size)

    def on_header_move(self, _section: int, old: int, new: int) -> None:
        self.moveSection(old, new)

    def on_model_update(self, top_left, bottom_right, _role) -> None:
        self.headerDataChanged(Qt.Horizontal, top_left.column(), bottom_right.column())

    def set_span(self, section: int, span: list):
        span += [section]  # protection to have section itself included into span
        for i in span:
            self._span[i] = section



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'term_deposit_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QHeaderView, QLabel,
    QLineEdit, QPushButton, QSizePolicy, QSpacerItem,
    QTableView, QWidget)

from jal.widgets.reference_selector import AccountSelector

class Ui_TermDepositOperation(object):
    def setupUi(self, TermDepositOperation):
        if not TermDepositOperation.objectName():
            TermDepositOperation.setObjectName(u"TermDepositOperation")
        TermDepositOperation.resize(969, 260)
        self.layout = QGridLayout(TermDepositOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.account_label = QLabel(TermDepositOperation)
        self.account_label.setObjectName(u"account_label")

        self.layout.addWidget(self.account_label, 2, 0, 1, 1)

        self.note = QLineEdit(TermDepositOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 3, 1, 1, 5)

        self.note_label = QLabel(TermDepositOperation)
        self.note_label.setObjectName(u"note_label")

        self.layout.addWidget(self.note_label, 3, 0, 1, 1)

        self.add_button = QPushButton(TermDepositOperation)
        self.add_button.setObjectName(u"add_button")

        self.layout.addWidget(self.add_button, 4, 4, 1, 1)

        self.del_button = QPushButton(TermDepositOperation)
        self.del_button.setObjectName(u"del_button")

        self.layout.addWidget(self.del_button, 4, 5, 1, 1)

        self.revert_button = QPushButton(TermDepositOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 9, 1, 1)

        self.main_label = QLabel(TermDepositOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 2)

        self.account_widget = AccountSelector(TermDepositOperation)
        self.account_widget.setObjectName(u"account_widget")

        self.layout.addWidget(self.account_widget, 2, 1, 1, 2)

        self.actions_table = QTableView(TermDepositOperation)
        self.actions_table.setObjectName(u"actions_table")
        self.actions_table.setAlternatingRowColors(True)
        self.actions_table.verticalHeader().setVisible(False)
        self.actions_table.verticalHeader().setMinimumSectionSize(20)
        self.actions_table.verticalHeader().setDefaultSectionSize(20)

        self.layout.addWidget(self.actions_table, 5, 0, 1, 6)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.layout.addItem(self.horizontalSpacer, 3, 6, 1, 1)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.layout.addItem(self.vertical_spacer, 6, 0, 1, 1)

        self.commit_button = QPushButton(TermDepositOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 8, 1, 1)

        self.details_label = QLabel(TermDepositOperation)
        self.details_label.setObjectName(u"details_label")

        self.layout.addWidget(self.details_label, 4, 0, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.layout.addItem(self.horizontalSpacer_2, 4, 3, 1, 1)


        self.retranslateUi(TermDepositOperation)

        QMetaObject.connectSlotsByName(TermDepositOperation)
    # setupUi

    def retranslateUi(self, TermDepositOperation):
        TermDepositOperation.setWindowTitle(QCoreApplication.translate("TermDepositOperation", u"Form", None))
        self.account_label.setText(QCoreApplication.translate("TermDepositOperation", u"Account", None))
        self.note_label.setText(QCoreApplication.translate("TermDepositOperation", u"Note", None))
#if QT_CONFIG(tooltip)
        self.add_button.setToolTip(QCoreApplication.translate("TermDepositOperation", u"Add activity", None))
#endif // QT_CONFIG(tooltip)
        self.add_button.setText("")
#if QT_CONFIG(tooltip)
        self.del_button.setToolTip(QCoreApplication.translate("TermDepositOperation", u"Remove activity", None))
#endif // QT_CONFIG(tooltip)
        self.del_button.setText("")
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("TermDepositOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
        self.main_label.setText(QCoreApplication.translate("TermDepositOperation", u"Term Deposit", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("TermDepositOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
        self.details_label.setText(QCoreApplication.translate("TermDepositOperation", u"Deposit activity", None))
    # retranslateUi




import logging
import os
import re
from enum import auto
from collections import UserDict
from PySide6.QtGui import QIcon, QPixmap
from jal.db.settings import JalSettings
from jal.db.tag import JalTag


ICON_PREFIX = "ui_"
FLAG_PREFIX = "flag_"
AUX_PREFIX = "aux_"

class JalIcon(UserDict):
    NONE = auto()
    ADD = auto()
    ADD_CHILD = auto()
    APP_MAIN = auto()
    BOND_AMORTIZATION = auto()
    BOND_INTEREST = auto()
    BUY = auto()
    CANCEL = auto()
    CHART = auto()
    CLEAN = auto()
    COPY = auto()
    DELISTING = auto()
    DEPOSIT_ACCOUNT = auto()
    DEPOSIT_OPEN = auto()
    DEPOSIT_CLOSE = auto()
    DETAILS = auto()
    DIVIDEND = auto()
    FEE = auto()
    INTEREST = auto()
    LIST = auto()
    MERGER = auto()
    MINUS = auto()
    OK = auto()
    PLUS = auto()
    REMOVE = auto()
    SELL = auto()
    SPINOFF = auto()
    SPLIT = auto()
    STOCK_DIVIDEND = auto()
    STOCK_VESTING = auto()
    SYMBOL_CHANGE = auto()
    TAG = auto()
    TAX = auto()
    TOTAL = auto()
    TRANSFER_IN = auto()
    TRANSFER_OUT = auto()
    TRANSFER_ASSET_IN = auto()
    TRANSFER_ASSET_OUT = auto()

    FLAG_PT = auto()
    FLAG_RU = auto()
    FLAG_US = auto()

    _icon_files = {
        ADD: "add.ico",
        ADD_CHILD: "add_child.ico",
        APP_MAIN: "jal.png",
        BOND_AMORTIZATION: "amortization.ico",
        BOND_INTEREST: "coupon.ico",
        BUY: "buy.ico",
        CANCEL: "cancel.ico",
        CHART: "chart.ico",
        CLEAN: "clean.ico",
        COPY: "copy.ico",
        DELISTING: "delisting.ico",
        DEPOSIT_ACCOUNT: "deposit_account.ico",
        DEPOSIT_OPEN: "deposit_open.ico",
        DEPOSIT_CLOSE: "deposit_close.ico",
        DETAILS: "details.ico",
        DIVIDEND: "dividend.ico",
        FEE: "fee.ico",
        INTEREST: "interest.ico",
        LIST: "list.ico",
        MERGER: "merger.ico",
        MINUS: "minus.ico",
        OK: "ok.ico",
        PLUS: "plus.ico",
        REMOVE: "remove.ico",
        SELL: "sell.ico",
        SPINOFF: "spinoff.ico",
        SPLIT: "split.ico",
        STOCK_DIVIDEND: "dividend_stock.ico",
        STOCK_VESTING: "vesting.ico",
        SYMBOL_CHANGE: "renaming.ico",
        TAG: "tag.ico",
        TAX: "tax.ico",
        TOTAL: "total.ico",
        TRANSFER_IN: "transfer_in.ico",
        TRANSFER_OUT: "transfer_out.ico",
        TRANSFER_ASSET_IN: "transfer_asset_in.ico",
        TRANSFER_ASSET_OUT: "transfer_asset_out.ico"
    }
    _icons = {}
    _flag_files = {
        FLAG_PT: "pt.png",
        FLAG_RU: "ru.png",
        FLAG_US: "en.png"
    }
    _flags = {
        'pt': FLAG_PT,
        'ru': FLAG_RU,
        'en': FLAG_US
    }

    # initiates class loading all icons listed in self._icon_files from given directory img_path (should and
    # with a system directory separator)
    def __init__(self):
        super().__init__()
        if self._icons:     # Already loaded - nothing to do
            return
        img_path = JalSettings.path(JalSettings.PATH_ICONS)
        for icon_id, filename in self._icon_files.items():
            self._icons[icon_id] = self.add_disabled_state(self.load_icon(img_path + ICON_PREFIX + filename))
        for icon_id, filename in self._flag_files.items():
            self._icons[icon_id] = self.load_icon(img_path + FLAG_PREFIX + filename)
        for filename in os.listdir(img_path):
            match = re.match(f"^{AUX_PREFIX}.*", filename)
            if match:
                self._icons[filename] = self.load_icon(img_path + filename)
        for tag_id, filename in JalTag.icon_files().items():
            self._icons[filename] = self.add_disabled_state(self.load_icon(img_path + filename))

    @staticmethod
    def load_icon(path) -> QIcon:
        icon = QIcon(path)
        if icon.isNull():
            logging.warning(f"Image file {path} not found")  # This error won't come to GUI as LogViewer is initialized later
        return icon

    @classmethod
    def __class_getitem__(cls, key) -> QIcon:
        return cls._icons.get(key, QIcon())

    @classmethod
    def country_flag(cls, country_code) -> QIcon:
        if country_code not in cls._flags:
            return QIcon()
        return cls._icons[cls._flags[country_code]]

    @classmethod
    def aux_icon(cls, icon_name) -> QIcon:
        filename = AUX_PREFIX + icon_name
        return cls._icons.get(filename, QIcon())

    # Iterates through all available images and creates a copy of images with adjusted alpha-channel (20% of initial value)
    # This new image is added to the icon as disabled state image
    def add_disabled_state(self, icon: QIcon) -> QIcon:
        disabled_icons = []
        for size in icon.availableSizes():
            icon_image = icon.pixmap(size).toImage()
            for y in range(icon_image.height()):
                for x in range(icon_image.width()):
                    pixel_color = icon_image.pixelColor(x, y)
                    pixel_color.setAlpha(pixel_color.alpha() / 5)
                    icon_image.setPixelColor(x, y, pixel_color)
            disabled_icons.append(QPixmap.fromImage(icon_image))
        for disabled_image in disabled_icons:
            icon.addPixmap(disabled_image, mode=QIcon.Mode.Disabled)
        return icon



import logging
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QStringListModel, QByteArray
from PySide6.QtWidgets import QHeaderView
from PySide6.QtSql import QSqlTableModel
from PySide6.QtGui import QFont
from jal.ui.widgets.ui_corporate_action_operation import Ui_CorporateActionOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.icons import JalIcon
from jal.widgets.delegates import WidgetMapperDelegateBase, AssetSelectorDelegate, FloatDelegate
from jal.db.view_model import JalViewModel
from jal.db.helpers import localize_decimal, now_ts
from jal.db.operations import LedgerTransaction


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate,
                          'asset_id': self.symbol_delegate,
                          'qty': self.decimal_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_CorporateActionOperation)
        self.operation_type = LedgerTransaction.CorporateAction
        self.combo_model = None

        self.asset_delegate = AssetSelectorDelegate()
        self.float_delegate = FloatDelegate(2)
        self.percent_delegate = FloatDelegate(2, percent=True)

        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.add_button.setIcon(JalIcon[JalIcon.ADD])
        self.ui.del_button.setIcon(JalIcon[JalIcon.REMOVE])
        self.ui.results_table.horizontalHeader().setFont(self.bold_font)
        self.ui.arrow.setText(" 🡆 ")  # it crashes if added via Qt-Designer

        self.ui.add_button.clicked.connect(self.addResult)
        self.ui.del_button.clicked.connect(self.delResult)

        super()._init_db("asset_actions")
        self.combo_model = QStringListModel([self.tr("N/A"),
                                             self.tr("Merger"),
                                             self.tr("Spin-Off"),
                                             self.tr("Symbol change"),
                                             self.tr("Split"),
                                             self.tr("Delisting")])
        self.ui.type.setModel(self.combo_model)

        self.mapper.setItemDelegate(CorporateActionWidgetDelegate(self.mapper))

        self.results_model = ResultsModel(self.ui.results_table, "action_results")
        self.results_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.ui.results_table.setModel(self.results_model)

        self.results_model.dataChanged.connect(self.onDataChange)
        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.asset_widget.changed.connect(self.mapper.submit)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.asset_widget, self.model.fieldIndex("asset_id"))
        self.mapper.addMapping(self.ui.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.ui.qty_edit, self.model.fieldIndex("qty"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))
        self.mapper.addMapping(self.ui.type, self.model.fieldIndex("type"), QByteArray().setRawData("currentIndex", 12))

        self.ui.results_table.setItemDelegateForColumn(2, self.asset_delegate)
        self.ui.results_table.setItemDelegateForColumn(3, self.float_delegate)
        self.ui.results_table.setItemDelegateForColumn(4, self.percent_delegate)

        self.model.select()
        self.results_model.select()
        self.results_model.configureView()

    def set_id(self, oid):
        self.results_model.setFilter(f"action_results.action_id = {oid}")  # First we need to select right children
        super().set_id(oid)

    @Slot()
    def addResult(self):
        new_record = self.results_model.record()
        new_record.setValue("qty", 0)
        new_record.setValue("value_share", 0)
        if not self.results_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.results_model.lastError().text())
            return

    @Slot()
    def delResult(self):
        selection = self.ui.results_table.selectionModel().selection().indexes()
        for idx in selection:
            self.results_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    def _save(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            oid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if oid is None:  # we just have saved new action record and need last inserted id
                oid = self.model.last_insert_id()
            for row in range(self.results_model.rowCount()):
                self.results_model.setData(self.results_model.index(row, self.results_model.fieldIndex("action_id")), oid)
            if not self.results_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.results_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.results_model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.results_model.setFilter(f"action_results.action_id = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", now_ts())
        new_record.setValue("number", '')
        new_record.setValue("account_id", account_id)
        new_record.setValue("type", 0)
        new_record.setValue("asset_id", 0)
        new_record.setValue("qty", '0')
        new_record.setValue("note", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.results_model.rowCount()):
            child_records.append(self.results_model.record(row))
        self.results_model.setFilter(f"action_results.action_id = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("action_id")
            assert self.results_model.insertRows(0, 1)
            self.results_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", now_ts())
        new_record.setValue("number", '')
        return new_record


# ----------------------------------------------------------------------------------------------------------------------
class ResultsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "action_id", self.tr("Asset"), self.tr("Qty"), self.tr("Share, %")]

    def footerData(self, section, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 2:
                return self.tr("Total")
            elif section == 4:
                total = Decimal('0')
                for row in range(self.rowCount()):
                    try:
                        value = Decimal(self.index(row, section).data())
                    except:
                        value = Decimal('0')
                    total += value
                return localize_decimal(total, precision=2, percent=True)
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 2:
                return Qt.AlignLeft | Qt.AlignVCenter
            else:
                return Qt.AlignRight | Qt.AlignVCenter
        return None

    def configureView(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 100)
        self._view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)



from decimal import Decimal
from functools import partial
from PySide6.QtCore import Qt, Slot, QObject, QAbstractTableModel
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QHeaderView
from jal.reports.reports import Reports
from jal.db.operations import AssetPayment
from jal.ui.reports.ui_assets_payments_report import Ui_AssetsPaymentsReportWidget
from jal.widgets.delegates import FloatDelegate
from jal.widgets.mdi import MdiWidget
from jal.widgets.helpers import ts2dt

JAL_REPORT_CLASS = "AssetsPaymentsReport"

# ----------------------------------------------------------------------------------------------------------------------
class AssetsPaymentsModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [self.tr("Date"), self.tr("Symbol"), self.tr("Asset"), self.tr("Type"),
                         self.tr("Amount"), self.tr("Tax"), self.tr("Note")]
        self._types = [self.tr("N/A"), self.tr("Dividend"), self.tr("Bond Interest"), self.tr("Stock Dividend"),
                       self.tr("Stock Vesting"), self.tr("Bond Amortization")]
        self._view = parent_view
        self._data = []
        self._account_id = 0
        self._begin = self._end = 0
        self._float_delegate = None
        self._total = Decimal('0')
        self._total_tax = Decimal('0')

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def headerWidth(self, section):
        return self._view.horizontalHeader().sectionSize(section)

    def data(self, index, role=Qt.DisplayRole, field=''):
        if role == Qt.DisplayRole:
            dividend = self._data[index.row()]
            return self.data_text(dividend, index.column())

    def data_text(self, dividend, column):
        if column == 0:
            return ts2dt(dividend.timestamp())
        if column == 1:
            return dividend.asset().symbol()
        if column == 2:
            return dividend.asset().name()
        if column == 3:
            return self._types[dividend.subtype()]
        if column == 4:
            return dividend.amount()
        if column == 5:
            return dividend.tax()
        if column == 6:
            return dividend.note()
        return ''

    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 4:
                return self._total
            if section == 5:
                return self._total_tax
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 4 or section == 5:
                return Qt.AlignRight | Qt.AlignVCenter
            else:
                return Qt.AlignLeft | Qt.AlignVCenter
        return None

    def updateView(self, account_id, dates):
        update = False
        if self._account_id != account_id:
            self._account_id = account_id
            update = True
        if self._begin != dates[0]:
            self._begin = dates[0]
            update = True
        if self._end != dates[1]:
            self._end = dates[1]
            update = True
        if update:
            self.prepareData()
            self.configureView()

    def prepareData(self):
        dividends = AssetPayment.get_list(self._account_id)
        self._data = [x for x in dividends if self._begin <= x.timestamp() <= self._end]
        self._total = sum([x.amount() for x in self._data])
        self._total_tax = sum([x.tax() for x in self._data])
        self.modelReset.emit()

    def configureView(self):
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._float_delegate = FloatDelegate(2, allow_tail=False)
        self._view.setItemDelegateForColumn(4, self._float_delegate)
        self._view.setItemDelegateForColumn(5, self._float_delegate)
        self._view.setColumnWidth(0, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        self._view.setColumnWidth(2, 200)
        if self._view.horizontalHeader().count():  # Next line crashes if there are no columns (count==0)
            self._view.horizontalHeader().setSectionResizeMode(6, QHeaderView.Stretch)

# ----------------------------------------------------------------------------------------------------------------------
class AssetsPaymentsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Assets' Payments")
        self.window_class = "AssetsPaymentsReportWindow"

# ----------------------------------------------------------------------------------------------------------------------
class AssetsPaymentsReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_AssetsPaymentsReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Assets' Payments")

        self.payments_model = AssetsPaymentsModel(self.ui.ReportTableView)
        self.ui.ReportTableView.setModel(self.payments_model)

        self.connect_signals_and_slots()

    def connect_signals_and_slots(self):
        self.ui.ReportAccountButton.changed.connect(self.updateReport)
        self.ui.ReportRange.changed.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTableView.model()))

    @Slot()
    def updateReport(self):
        self.ui.ReportTableView.model().updateView(account_id=self.ui.ReportAccountButton.account_id,
                                                   dates=self.ui.ReportRange.getRange())



import os
from enum import auto
from jal.db.db import JalDB
from PySide6.QtCore import QStandardPaths, QFileInfo
from jal.constants import Setup

class FolderFor:
    Statement = auto()
    Report = auto()

class JalSettings(JalDB):
    __RECENT_PREFIX = "RecentFolder_"
    __folders = {
        FolderFor.Statement: "Statement",
        FolderFor.Report: "Report"
    }

    def __init__(self):
        super().__init__()

    @staticmethod
    def path(path_type) -> str:
        app_path = JalDB.get_app_path()
        if path_type == JalDB.PATH_APP:
            return app_path
        if path_type == JalDB.PATH_DB_FILE:
            return JalDB.get_db_path()
        if path_type == JalDB.PATH_LANG:
            return app_path + Setup.LANG_PATH + os.sep
        if path_type == JalDB.PATH_ICONS:
            return app_path + Setup.ICONS_PATH + os.sep
        if path_type == JalDB.PATH_LANG_FILE:
            return app_path + Setup.LANG_PATH + os.sep + JalSettings().getLanguage() + '.qm'
        if path_type == JalDB.PATH_TAX_REPORT_TEMPLATE:
            return app_path + Setup.EXPORT_PATH + os.sep + Setup.TAX_REPORT_PATH + os.sep
        if path_type == JalDB.PATH_TEMPLATES:
            return app_path + Setup.EXPORT_PATH + os.sep + Setup.TEMPLATE_PATH + os.sep

    def getValue(self, key, default=None):
        value = self._read("SELECT value FROM settings WHERE name=:key", [(":key", key)])
        if value is None:
            value = default
        return value

    def setValue(self, key, value):
        self._exec("INSERT OR REPLACE INTO settings(id, name, value) "
                   "VALUES((SELECT id FROM settings WHERE name=:key), :key, :value)",
                   [(":key", key), (":value", value)], commit=True)

    # Returns 2-letter language code that corresponds to current 'Language' settings in DB
    def getLanguage(self):
        lang_id = self.getValue('Language', default=1)
        return self._read("SELECT language FROM languages WHERE id = :language_id", [(':language_id', lang_id)])

    # Set 'Language' setting in DB that corresponds to given 2-letter language code
    def setLanguage(self, language_code):
        lang_id = self._read("SELECT id FROM languages WHERE language = :language_code",
                             [(':language_code', language_code)])
        self.setValue('Language', lang_id)

    def getRecentFolder(self, folder_type: int, default: str=''):
        folder = self.getValue(self.__RECENT_PREFIX + self.__folders[folder_type])
        if not folder: folder = QStandardPaths.writableLocation(QStandardPaths.DocumentsLocation)
        if not folder: folder = default
        return folder

    def setRecentFolder(self, folder_type: int, folder: str):
        file_info = QFileInfo(folder)
        if file_info.isDir():
            path = file_info.absoluteFilePath()
        else:
            path = file_info.absolutePath()
        self.setValue(self.__RECENT_PREFIX + self.__folders[folder_type], path)



from math import floor, ceil
from decimal import Decimal

from PySide6.QtCore import Qt, QMargins, QDateTime
from PySide6.QtWidgets import QWidget, QHBoxLayout
from PySide6.QtCharts import QChartView, QLineSeries, QScatterSeries, QDateTimeAxis, QValueAxis, QXYSeries
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.operations import LedgerTransaction, Transfer, CorporateAction
from jal.constants import CustomColor
from jal.widgets.mdi import MdiWidget
from jal.widgets.helpers import ts2d


class ChartWidget(QWidget):
    def __init__(self, parent, quotes, trades, data_range, currency_name):
        super().__init__(parent=parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)

        self.quotes_series = QLineSeries()
        for point in quotes:            # Conversion to 'float' in order not to get 'int' overflow on some platforms
            self.quotes_series.append(float(point['timestamp']), point['quote'])
        self.quotes_series.setColor(CustomColor.DarkBlue)

        self._trades = trades
        self.trade_series = QScatterSeries()
        points_config = {}
        for i, point in enumerate(trades):            # Conversion to 'float' in order not to get 'int' overflow on some platforms
            self.trade_series.append(float(point['timestamp']), point['price'])
            points_config[i] = {QXYSeries.PointConfiguration.Color: point['color']}
        self.trade_series.setMarkerSize(7)
        self.trade_series.setBorderColor(CustomColor.Grey)
        self.trade_series.setPointsConfiguration(points_config)
        self.trade_series.hovered.connect(self.MouseOverTrade)

        self.axisX = QDateTimeAxis()
        self.axisX.setTickCount(11)
        self.axisX.setRange(QDateTime().fromSecsSinceEpoch(data_range[0]), QDateTime().fromSecsSinceEpoch(data_range[1]))
        self.axisX.setFormat("yyyy/MM/dd")
        self.axisX.setLabelsAngle(-90)
        self.axisX.setTitleText("Date")

        self.axisY = QValueAxis()
        self.axisY.setTickCount(11)
        self.axisY.setRange(data_range[2], data_range[3])
        self.axisY.setTitleText("Price, " + currency_name)

        self.chartView = QChartView()
        self.chartView.chart().addSeries(self.quotes_series)
        self.chartView.chart().addSeries(self.trade_series)
        self.chartView.chart().addAxis(self.axisX, Qt.AlignBottom)
        self.quotes_series.attachAxis(self.axisX)
        self.trade_series.attachAxis(self.axisX)
        self.chartView.chart().addAxis(self.axisY, Qt.AlignLeft)
        self.quotes_series.attachAxis(self.axisY)
        self.trade_series.attachAxis(self.axisY)
        self.chartView.chart().legend().hide()
        self.chartView.setViewportMargins(0, 0, 0, 0)
        self.chartView.chart().layout().setContentsMargins(0, 0, 0, 0)  # To remove extra spacing around chart
        self.chartView.chart().setBackgroundRoundness(0)  # To remove corner rounding
        self.chartView.chart().setMargins(QMargins(0, 0, 0, 0))  # Allow chart to fill all space

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove extra space around layout
        self.layout.addWidget(self.chartView)
        self.setLayout(self.layout)

    def MouseOverTrade(self, point, state):
        if state:
            trade = [x for x in self._trades if float(x['timestamp']) == point.x() and float(x['price']) == point.y()]
            qty = sum([x['qty'] for x in trade])
            avg_price = sum([x['price']*x['qty'] for x in trade]) / qty
            tip_text = ts2d(int(point.x()/1000)) + ": "
            tip_text += f"+{qty}@{avg_price}" if qty > 0 else f"{qty}@{avg_price}"
            if len(trade) and trade[0]['text']:
                tip_text += "\n" + trade[0]['text']
            self.setToolTip(tip_text)
        else:
            self.setToolTip("")

class ChartWindow(MdiWidget):
    def __init__(self, account_id, asset_id, currency_id, timestamp, parent=None):
        super().__init__(parent)

        self.account_id = account_id
        self.asset_id = asset_id
        self.currency_id = currency_id if asset_id != currency_id else 1  # Check whether we have currency or asset
        self.asset_name = JalAsset(self.asset_id).symbol(JalAccount(self.account_id).currency())
        self.quotes = []
        self.trades = []
        self.currency_name = ''
        self.range = [0, 0, 0, 0]

        self.prepare_chart_data(timestamp)

        self.chart = ChartWidget(self, self.quotes, self.trades, self.range, self.currency_name)

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove extra space around layout
        self.layout.addWidget(self.chart)
        self.setLayout(self.layout)

        self.setWindowTitle(self.tr("Price chart for ") + self.asset_name)

        self.ready = True

    def load_open_trades(self, account, asset, end_time):
        trades = []
        positions = account.open_trades_list(asset, end_time)
        for trade in positions:
            marker_color = CustomColor.LightYellow
            text = ''
            operation = trade['operation']
            if operation.type() == LedgerTransaction.Trade:
                if trade['remaining_qty'] >= 0:
                    marker_color = CustomColor.LightGreen
                    text = self.tr("Buy")
                else:
                    marker_color = CustomColor.LightRed
                    text = self.tr("Sell")
            if operation.type() == LedgerTransaction.Transfer:
                transfer_out = LedgerTransaction().get_operation(operation.type(), operation.id(), Transfer.Outgoing)
                trades += self.load_open_trades(transfer_out.account(), asset, transfer_out.timestamp()-1)  # get position just before the transfer
                text = self.tr("Transfer ") + f"{transfer_out.account_name()}"
            if operation.type() == LedgerTransaction.CorporateAction and operation.subtype() == CorporateAction.Split:
                trades += self.load_open_trades(account, asset, operation.timestamp()-1)  # get position just before the split
                text = operation.description().split('\n')[0]
            if operation.type() == LedgerTransaction.AssetPayment:
                text = operation.name() + "\n" + operation.description().split('\n')[0]
            trades.append({
                'timestamp': operation.timestamp() * 1000,  # timestamp to ms
                'price': trade['price'],
                'qty': trade['remaining_qty'],
                'color': marker_color,
                'text': text
            })
        return trades

    def prepare_chart_data(self, end_time):
        account = JalAccount(self.account_id)
        asset = JalAsset(self.asset_id)
        self.currency_name = JalAsset(account.currency()).symbol()
        self.trades = self.load_open_trades(account, asset, end_time)
        start_time = 0 if not self.trades else min([x['timestamp'] for x in self.trades])/1000 - 2592000  # Shift back by 30 days
        quotes = asset.quotes(start_time, end_time, self.currency_id)
        for quote in quotes:
            self.quotes.append({'timestamp': quote[0] * 1000, 'quote': quote[1]})  # timestamp to ms
        if self.quotes or self.trades:
            min_price = min([x['quote'] for x in self.quotes] + [x['price'] for x in self.trades])
            max_price = max([x['quote'] for x in self.quotes] + [x['price'] for x in self.trades])
            min_ts = min([x['timestamp'] for x in self.quotes] + [x['timestamp'] for x in self.trades]) / 1000
            max_ts = max([x['timestamp'] for x in self.quotes] + [x['timestamp'] for x in self.trades]) / 1000
        else:
            self.range = [0, 0, 0, 0]
            return
        # push range apart if we have very close points
        if min_price == max_price:
            min_price = Decimal('0.95') * min_price
            max_price = Decimal('1.05') * max_price
        if min_ts == max_ts:
            min_ts = 0.95 * min_ts
            max_ts = 1.05 * max_ts
        # Round min/max values to near "round" values in order to have 10 nice intervals
        step = Decimal(10) ** Decimal(floor(Decimal.log10(max_price - min_price)))
        min_price = floor(min_price / step) * step
        max_price = ceil(max_price / step) * step
        # Add a gap at the beginning and end
        min_ts = int(min_ts - 86400 * 3)
        max_ts = int(max_ts + 86400 * 3)
        self.range = [min_ts, max_ts, min_price, max_price]



import logging
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QByteArray
from PySide6.QtWidgets import QMessageBox, QHeaderView
from PySide6.QtSql import QSqlTableModel
from PySide6.QtGui import QFont
from jal.ui.widgets.ui_income_spending_operation import Ui_IncomeSpendingOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.icons import JalIcon
from jal.db.view_model import JalViewModel
from jal.db.helpers import localize_decimal, db_row2dict, now_ts
from jal.db.operations import LedgerTransaction
from jal.widgets.delegates import WidgetMapperDelegateBase, FloatDelegate, CategorySelectorDelegate, TagSelectorDelegate


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_IncomeSpendingOperation)
        self.operation_type = LedgerTransaction.IncomeSpending
        super()._init_db("actions")

        self.category_delegate = CategorySelectorDelegate()
        self.tag_delegate = TagSelectorDelegate()
        self.float_delegate = FloatDelegate(2)

        self.ui.timestamp_editor.setFixedWidth(self.ui.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.a_currency.setText(self.tr("Paid in foreign currency:"))
        self.ui.details_table.horizontalHeader().setFont(self.bold_font)
        self.ui.add_button.setIcon(JalIcon[JalIcon.ADD])
        self.ui.del_button.setIcon(JalIcon[JalIcon.REMOVE])
        self.ui.copy_button.setIcon(JalIcon[JalIcon.COPY])

        self.ui.add_button.clicked.connect(self.add_child)
        self.ui.copy_button.clicked.connect(self.copy_child)
        self.ui.del_button.clicked.connect(self.delete_child)
        self.model.beforeInsert.connect(self.before_record_insert)
        self.model.beforeUpdate.connect(self.before_record_update)

        self.mapper.setItemDelegate(IncomeSpendingWidgetDelegate(self.mapper))

        self.details_model = DetailsModel(self.ui.details_table, "action_details")
        self.details_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.ui.details_table.setModel(self.details_model)
        self.details_model.dataChanged.connect(self.onDataChange)

        self.ui.account_widget.changed.connect(self.mapper.submit)
        self.ui.peer_widget.changed.connect(self.mapper.submit)
        self.ui.a_currency.changed.connect(self.mapper.submit)
        self.ui.a_currency.name_updated.connect(self.details_model.set_alternative_currency)

        self.mapper.addMapping(self.ui.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.currency, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.peer_widget, self.model.fieldIndex("peer_id"))
        self.mapper.addMapping(self.ui.a_currency, self.model.fieldIndex("alt_currency_id"), QByteArray("currency_id_str"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.ui.details_table.setItemDelegateForColumn(2, self.category_delegate)
        self.ui.details_table.setItemDelegateForColumn(3, self.tag_delegate)
        self.ui.details_table.setItemDelegateForColumn(4, self.float_delegate)
        self.ui.details_table.setItemDelegateForColumn(5, self.float_delegate)

        self.model.select()
        self.details_model.select()
        self.details_model.configure_view()

    def set_id(self, oid):
        self.details_model.setFilter(f"action_details.pid = {oid}")  # First we need to select right children
        super().set_id(oid)

    @Slot()
    def add_child(self):
        new_record = self.details_model.record()
        new_record.setNull("tag_id")
        new_record.setValue("amount", '0')
        new_record.setValue("amount_alt", '0')
        if not self.details_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.details_model.lastError().text())
            return

    @Slot()
    def copy_child(self):
        idx = self.ui.details_table.selectionModel().selection().indexes()
        src_record = self.details_model.record(idx[0].row())
        new_record = self.details_model.record()
        new_record.setValue("category_id", src_record.value("category_id"))
        if src_record.value("tag_id"):
            new_record.setValue("tag_id", src_record.value("tag_id"))
        else:
            new_record.setNull("tag_id")
        new_record.setValue("amount", src_record.value("amount"))
        new_record.setValue("amount_alt", src_record.value("amount_alt"))
        new_record.setValue("note", src_record.value("note"))
        if not self.details_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.details_model.lastError().text())
            return

    @Slot()
    def delete_child(self):
        selection = self.ui.details_table.selectionModel().selection().indexes()
        for idx in selection:
            self.details_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    def _validated(self):
        fields = db_row2dict(self.model, 0)
        if not fields['account_id'] or not fields['peer_id']:
            return False
        if not self.details_model.rowCount():
            QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Operation contains no details"), QMessageBox.Ok)
            return False
        for row in range(self.details_model.rowCount()):
            fields = db_row2dict(self.details_model, row)
            if fields['category_id'] is None or fields['category_id'] == 0:
                QMessageBox().warning(self, self.tr("Incomplete data"),
                                      self.tr("Category isn't set for '{}' (Amount: {})".format(fields['note'], fields['amount'])), QMessageBox.Ok)
                return False
        return True

    def _save(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            pid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if pid is None:  # we just have saved new action record and need last inserted id
                pid = self.model.last_insert_id()
            for row in range(self.details_model.rowCount()):   # Set PID for all child records
                self.details_model.setData(self.details_model.index(row, self.details_model.fieldIndex("pid")), pid)
            if not self.details_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.details_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.model.database().commit()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.details_model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.details_model.setFilter(f"action_details.pid = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", now_ts())
        new_record.setValue("account_id", account_id)
        new_record.setValue("peer_id", 0)
        new_record.setValue("alt_currency_id", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.details_model.rowCount()):
            child_records.append(self.details_model.record(row))
        self.details_model.setFilter(f"action_details.pid = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("pid")
            if not record.value("tag_id"):
                record.setNull("tag_id")
            assert self.details_model.insertRows(0, 1)
            self.details_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", now_ts())
        return new_record

    def before_record_insert(self, record):
        if record.value("alt_currency_id") == 0 or record.value("alt_currency_id") == '':
            record.setNull("alt_currency_id")

    def before_record_update(self, _row, record):
        self.before_record_insert(record)   # processing is the same as before insert


# ----------------------------------------------------------------------------------------------------------------------
class DetailsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "pid", self.tr("Category"), self.tr("Tag"),
                         self.tr("Amount"), self.tr("Amount"), self.tr("Note")]
        self.alt_currency_name = ''

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if section == 5 and orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section] + ', ' + self.alt_currency_name
        else:
            return super().headerData(section, orientation, role)

    def footerData(self, section, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 6:
                return self.tr("Total")
            elif section == 4 or section == 5:
                total = Decimal('0')
                for row in range(self.rowCount()):
                    try:
                        value = Decimal(self.index(row, section).data())
                    except:
                        value = Decimal('0')
                    total += value
                return localize_decimal(total, precision=2)
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section == 6:
                return Qt.AlignLeft | Qt.AlignVCenter
            else:
                return Qt.AlignRight | Qt.AlignVCenter
        return None

    def configure_view(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnHidden(5, True)
        self._view.setColumnWidth(2, 200)
        self._view.setColumnWidth(3, 200)
        self._view.setColumnWidth(4, 100)
        self._view.setColumnWidth(5, 100)
        self._view.horizontalHeader().setSectionResizeMode(6, QHeaderView.Stretch)
        self._view.horizontalHeader().moveSection(6, 0)

    def set_alternative_currency(self, currency_name):
        if currency_name:
            self._view.setColumnHidden(5, False)
            self.alt_currency_name = currency_name
            self.headerDataChanged.emit(Qt.Horizontal, 5, 5)
        else:
            for row in range(self.rowCount()):
                self.setData(self.index(row, self.fieldIndex("amount_alt")), '0')  # Reset all alternative amounts
            self._view.setColumnHidden(5, True)



from __future__ import annotations
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QDate
from PySide6.QtGui import QBrush, QFont
from PySide6.QtWidgets import QHeaderView
from jal.constants import CustomColor
from jal.db.helpers import localize_decimal
from jal.db.tree_model import AbstractTreeItem, ReportTreeModel
from jal.db.asset import JalAsset
from jal.db.account import JalAccount
from jal.db.deposit import JalDeposit
from jal.widgets.delegates import GridLinesDelegate, FloatDelegate
from jal.widgets.icons import JalIcon


# ----------------------------------------------------------------------------------------------------------------------
class AccountTreeItem(AbstractTreeItem):
    def __init__(self, data=None, parent=None, group=''):
        super().__init__(parent, group)
        if data is None:
            self._data = {
                "account_tag": '', "icon_id": JalIcon.NONE, "account": 0, "account_name": '', "currency": 0, "currency_name": '',
                "value": Decimal('0'), "value_common": Decimal('0'), "reconciled": 0, "active": 1, "font": "bold"
            }
        else:
            self._data = data.copy()

    def _calculateGroupTotals(self, child_data):
        self._data['account_name'] = child_data['account_tag']
        self._data['icon_id'] = child_data['icon_id']
        self._data['value_common'] += child_data['value_common']

    def _afterParentGroupUpdate(self, group_data):
        pass

    def details(self):
        return self._data

    # assigns group value if this tree item is a group item
    def setGroupValue(self, value):
        if self._group:
            self._data[self._group] = value

    def getGroup(self):
        if self._group:
            return self._group, self._data[self._group]
        else:
            return None

    def getGroupLeaf(self, group_fields: list, item: AccountTreeItem) -> AccountTreeItem:  # FIXME: similar code in Holdings model - need to combine and simplify
        if group_fields:
            group_item = None
            group_name = group_fields[0]
            for child in self._children:
                if child.details()[group_name] == item.details()[group_name]:
                    group_item = child
            if group_item is None:
                group_item = AccountTreeItem(None, parent=self, group=group_name)
                group_item.setGroupValue(item.details()[group_name])
                self._children.append(group_item)
            return group_item.getGroupLeaf(group_fields[1:], item)
        else:
            return self


# ----------------------------------------------------------------------------------------------------------------------
class BalancesModel(ReportTreeModel):
    ACCOUNT_ROLE = Qt.UserRole

    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._grid_delegate = None
        self._float_delegate = None
        self._view = parent_view
        self._data = []
        self._currency = 0
        self._currency_name = ''
        self._active_only = True
        self._date = QDate.currentDate().endOfDay(Qt.UTC).toSecsSinceEpoch()
        self.bold_font = QFont()
        self.bold_font.setBold(True)
        self.italic_font = QFont()
        self.italic_font.setItalic(True)
        self._fonts = {'normal': None, 'bold': self.bold_font, 'italic': self.italic_font}
        self._columns = [
            {'name': self.tr("Account"), 'field': 'account_name'},
            {'name': self.tr("Balance"), 'field': 'value'},
            {'name': self.tr(" "), 'field': 'currency_name'},
            {'name': self.tr("Balance, "), 'field': 'value_common'}
        ]

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        value = super().headerData(section, orientation, role)
        if orientation == Qt.Horizontal and role == Qt.DisplayRole and section == self.fieldIndex('value_common'):
            value += self._currency_name
        return value

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        assert index.column() in range(len(self._columns))
        try:
            item = index.internalPointer()
            if role == Qt.DisplayRole:
                return item.details().get(self._columns[index.column()]['field'], None)
            if role == Qt.FontRole:
                return self._fonts.get(item.details()['font'], None)
            if role == Qt.BackgroundRole and index.column() == self.fieldIndex('value'):
                return self.data_background(item.details().get('unreconciled', 0), self._view.isEnabled())
            if role == Qt.DecorationRole and item.isGroup() and index.column() == self.fieldIndex('account_name'):
                return JalIcon[item.details().get('icon_id', JalIcon.NONE)]
            if role == self.ACCOUNT_ROLE:
                return item.details().get('account', 0)
            return None
        except Exception as e:
            print(e)

    def footerData(self, section, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == self.fieldIndex('account_name'):
                return self.tr("Total:")
            elif section == self.fieldIndex('value_common'):
                totals_data = self._root.details()
                return localize_decimal(totals_data[self._columns[section]['field']], precision=2)
        if role == Qt.FontRole:
            return self.bold_font
        if role == Qt.TextAlignmentRole:
            if section == self.fieldIndex('value_common'):
                return Qt.AlignRight | Qt.AlignVCenter
            else:
                return Qt.AlignLeft | Qt.AlignVCenter
        if role == Qt.DecorationRole and section == self.fieldIndex('account_name'):
            return JalIcon[JalIcon.TOTAL]
        return None

    def data_background(self, unreconciled, enabled=True):
        factor = 100 if enabled else 125
        if unreconciled > 15:
            return QBrush(CustomColor.LightRed.lighter(factor))
        if unreconciled > 7:
            return QBrush(CustomColor.LightYellow.lighter(factor))
        return None

    def configureView(self):
        for field in [x['field'] for x in self._columns]:
            if field == 'account_name':
                self._view.header().setSectionResizeMode(self.fieldIndex(field), QHeaderView.Stretch)
            else:
                self._view.header().setSectionResizeMode(self.fieldIndex(field), QHeaderView.ResizeToContents)
        self._grid_delegate = GridLinesDelegate(self._view)
        self._float_delegate = FloatDelegate(2, allow_tail=False, empty_zero=True, parent=self._view)
        self._view.setItemDelegateForColumn(self.fieldIndex('account_name'), self._grid_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('currency_name'), self._grid_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('value'), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('value_common'), self._float_delegate)
        self._view.footer().set_span(self.fieldIndex('account_name'), [self.fieldIndex('account_name'), self.fieldIndex('value')])
        self._view.footer().set_span(self.fieldIndex('value_common'), [self.fieldIndex('currency_name'), self.fieldIndex('value_common')])
        super().configureView()

    @Slot()
    def setCurrency(self, currency_id):
        if self._currency != currency_id:
            self._currency = currency_id
            self._currency_name = JalAsset(currency_id).symbol()
            self.prepareData()

    @Slot()
    def setDate(self, new_date):
        if self._date != new_date.endOfDay(Qt.UTC).toSecsSinceEpoch():
            self._date = new_date.endOfDay(Qt.UTC).toSecsSinceEpoch()
            self.prepareData()

    @Slot()
    def toggleActive(self, state):
        if state == 0:
            self._active_only = True
        else:
            self._active_only = False
        self.prepareData()

    def getAccountId(self, index):
        if not index.isValid():
            return 0
        return index.internalPointer().details().get('account', 0)

    def update(self):
        self.prepareData()

    # Populate table balances with data calculated for given parameters of model: _currency, _date, _active_only
    def prepareData(self):
        self.setGrouping("account_tag")
        balances = []
        accounts = JalAccount.get_all_accounts(active_only=self._active_only)
        for account in accounts:
            value = account.balance(self._date)
            rate = JalAsset(account.currency()).quote(self._date, self._currency)[1]
            if value != Decimal('0'):
                balances.append({
                    "account_tag": account.tag().name() if account.tag().name() else self.tr("Without tag"),
                    "icon_id": account.tag().icon(),
                    "account": account.id(),
                    "account_name": account.name(),
                    "currency": account.currency(),
                    "currency_name": JalAsset(account.currency()).symbol(),
                    "value": value,
                    "value_common": value * rate,
                    "unreconciled": (account.last_operation_date() - account.reconciled_at())/86400,
                    "active": account.is_active(),
                    "font": 'normal' if account.is_active() else 'italic'
                })
        for deposit in JalDeposit.get_term_deposits(self._date):
            rate = deposit.currency().quote(self._date, self._currency)[1]
            balances.append({
                "account_tag": self.tr("Term deposits"),
                "icon_id": JalIcon.DEPOSIT_ACCOUNT,
                "account": 0,
                "account_name": deposit.name(),
                "currency": deposit.currency(),
                "currency_name": deposit.currency().symbol(),
                "value": deposit.balance(self._date),
                "value_common": deposit.balance(self._date) * rate,
                "unreconciled": 0,
                "active": 1,
                "font": 'normal'
            })
        # Sort data items and add them into the tree in right order
        balances = sorted(balances, key=lambda x: (x['account_tag'], x['account_name']))
        self._root = AccountTreeItem()
        for position in balances:
            new_item = AccountTreeItem(position)
            leaf = self._root.getGroupLeaf(self._groups, new_item)
            leaf.appendChild(new_item)
        super().prepareData()



import os
import logging
from datetime import datetime
from jal.constants import CustomColor
from jal.widgets.icons import JalIcon
from PySide6.QtCore import Qt, Slot, qInstallMessageHandler, QtMsgType
from PySide6.QtWidgets import QApplication, QPlainTextEdit, QLabel, QPushButton
from PySide6.QtGui import QBrush


# Adapter class to have custom log handler that may be passed to logger.addHandler/logger.removeHandler methods and
# then forward all messages parent view to display them
class LogHandler(logging.Handler):
    def __init__(self, parent_view):
        self._parent_view = parent_view
        super().__init__()

    def emit(self, record, **kwargs):
        message = self.format(record)
        colors = {
            logging.DEBUG: CustomColor.Grey,
            logging.INFO: None,
            logging.WARNING: CustomColor.LightRed,
            logging.ERROR: CustomColor.LightRed,
            logging.CRITICAL: CustomColor.LightRed
        }
        try:
            message_color = colors[record.levelno]
        except KeyError:
            message_color = CustomColor.LightRed
        self._parent_view.displayMessage(message, message_color)


# A GUI class to display messages from python logging unit in a normal multi-line text area
class LogViewer(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()
        self._logger = None     # Here an instance of current logger will be stored
        self._log_handler = LogHandler(self)
        self.setReadOnly(True)
        self.status_bar = None    # Status bar where notifications and control are located
        self.expandButton = None  # Button that shows/hides log window
        self.notification = None  # Here is QLabel element to display LOG update status
        self.clear_color = None   # Variable to store initial "clear" background color
        self.collapsed_text = self.tr("▶ logs")
        self.expanded_text = self.tr("▲ logs")

        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.addAction(JalIcon[JalIcon.COPY], self.tr('Copy'), self._copy2clipboard)
        self.addAction(self.tr('Select all'), self.selectAll)
        self.addAction(JalIcon[JalIcon.CLEAN], self.tr('Clear'), self.clear)

    def _copy2clipboard(self):
        cursor = self.textCursor()
        text = cursor.selectedText() if cursor.selectedText() else self.toPlainText()
        QApplication.clipboard().setText(text)

    def startLogging(self):
        self._logger = logging.getLogger()
        self._logger.addHandler(self._log_handler)
        log_level = os.environ.get('LOGLEVEL', 'INFO').upper()
        self._logger.setLevel(log_level)
        self._log_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        qInstallMessageHandler(self.qtLogHandler)

    def qtLogHandler(self, mode, _context, message):
        colors = {
            QtMsgType.QtDebugMsg: CustomColor.Grey,
            QtMsgType.QtInfoMsg: None,
            QtMsgType.QtWarningMsg: CustomColor.LightRed,
            QtMsgType.QtCriticalMsg: CustomColor.LightRed,
            QtMsgType.QtFatalMsg: CustomColor.LightRed,
            QtMsgType.QtSystemMsg: CustomColor.LightRed
        }
        try:
            message_color = colors[mode]
        except KeyError:
            message_color = CustomColor.LightRed
        message = f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]} - Qt - {message}"
        self.displayMessage(message, message_color)

    def stopLogging(self):
        self._logger.removeHandler(self._log_handler)    # Removing handler (but it doesn't prevent exception at exit)
        logging.raiseExceptions = False                  # Silencing logging module exceptions

    def displayMessage(self, message: str, color: CustomColor = None):
        color = self.clear_color if color is None else color

        # Store message in log window
        text_format = self.currentCharFormat()
        text_format.setForeground(QBrush(color))
        self.setCurrentCharFormat(text_format)
        self.appendPlainText(message)

        # Show in status bar
        if self.notification:
            palette = self.notification.palette()
            palette.setColor(self.notification.foregroundRole(), color)
            self.notification.setPalette(palette)
            msg = message.replace('\n', "; ")  # Get rid of new lines in error message
            elided_text = self.notification.fontMetrics().elidedText(msg, Qt.ElideRight, self.get_available_width())
            self.notification.setText(elided_text)
        # Set button color
        if self.expandButton:
            palette = self.expandButton.palette()
            palette.setColor(self.expandButton.foregroundRole(), color)
        self.app.processEvents()

    def showEvent(self, event):
        self.cleanNotification()
        super().showEvent(event)

    def setStatusBar(self, status_bar):
        self.setVisible(False)
        self.status_bar = status_bar

        self.expandButton = QPushButton(self.collapsed_text, parent=self)
        self.expandButton.setFixedWidth(self.expandButton.fontMetrics().horizontalAdvance(self.collapsed_text) * 1.25)
        self.expandButton.setCheckable(True)
        self.expandButton.clicked.connect(self.showLogs)
        self.status_bar.addWidget(self.expandButton)

        self.notification = QLabel(self)
        self.status_bar.addWidget(self.notification)
        self.notification.setAutoFillBackground(True)
        self.clear_color = self.expandButton.palette().color(self.notification.foregroundRole())

    def removeStatusBar(self):
        self.cleanNotification()
        self.notification = None

    def cleanNotification(self):
        if self.notification:
            palette = self.notification.palette()
            palette.setColor(self.notification.foregroundRole(), self.clear_color)
            self.notification.setPalette(palette)
            self.notification.setText("")

    @Slot()
    def showLogs(self):
        self.setVisible(self.expandButton.isChecked())
        text = self.expanded_text if self.expandButton.isChecked() else self.collapsed_text
        self.expandButton.setText(text)

    # Calculates maximum width that is free on status bar
    def get_available_width(self):
        width = self.status_bar.width()
        for child in self.status_bar.children():
            if hasattr(child, "width") and child != self.notification:
                width -= child.width()
        return width - 8    # return calculated width reduced by small safety gap



from datetime import datetime, timezone, timedelta
from decimal import Decimal, InvalidOperation
from PySide6.QtCore import QLocale
from jal.constants import Setup, JalGlobals


# -------------------------------------------------------------------------------------------------------------------
# Return "canonical" string for decimal number
def format_decimal(d) -> str:
    return str(d.normalize())


# Removes exponent and trailing zeros from Decimal number
def remove_exponent(d) -> Decimal:
    return d.quantize(Decimal(1)) if d == d.to_integral() else d.normalize()


# Make a locale-specific string from a Decimal value rounded to 'precision' digits after decimal point
# Multiplies value by 100 if 'percent' is True
# Returns empty string for None value
# Returns Setup.NULL_VALUE for NaN value
def localize_decimal(value: Decimal, precision: int = None, percent: bool = False, sign: bool = False) -> str:
    if value is None:
        return ''
    if value.is_nan():
        return Setup.NULL_VALUE
    if percent:
        value *= Decimal('100')
    f_str = '{:'
    if sign:
        f_str += '+'
    f_str += ','
    if precision:
        f_str += f".{precision}"
    f_str += "f}"
    formatted_number = f_str.format(value)
    pos = formatted_number.find('.')
    if pos==0:
        formatted_number = '0' + formatted_number
        pos += 1
    if pos > 0:
        if precision is not None:
            formatted_number = formatted_number[:pos+precision+1]
        else:
            formatted_number = formatted_number.rstrip('0')
            formatted_number = formatted_number[:-1] if formatted_number[-1]=='.' else formatted_number
    formatted_number = formatted_number.replace(',', JalGlobals().number_group_separator).replace('.', JalGlobals().number_decimal_point)
    return formatted_number


# Make number not locale-specific - i.e. replace decimal separator with '.' and remove any thousand separators
# Divide value by 100 if 'percent' is True
def delocalize_decimal(value: str, percent: bool = False) -> Decimal:
    number_text = value.replace(' ', '')
    number_text = number_text.replace(QLocale().groupSeparator(), '')
    number_text = number_text.replace(QLocale().decimalPoint(), '.')
    try:
        number = Decimal(number_text)
    except InvalidOperation:
        number = Decimal('0')
    if percent:
        number /= Decimal('100')
    return number

# -------------------------------------------------------------------------------------------------------------------
# Return a row from the model in form of {"field_name": value} dictionary
def db_row2dict(model, row) -> dict:
    record = model.record(row)
    return {record.field(x).name(): record.value(x) for x in range(record.count())}

# -------------------------------------------------------------------------------------------------------------------
# Returns timestamp of the first second of the year of given timestamp
def year_begin(timestamp: int) -> int:
    begin = datetime.utcfromtimestamp(timestamp).replace(month=1, day=1, hour=0, minute=0, second=0)
    return int(begin.replace(tzinfo=timezone.utc).timestamp())

# Returns timestamp of the last second of the year of given timestamp
def year_end(timestamp: int) -> int:
    end = datetime.utcfromtimestamp(timestamp).replace(month=1, day=1, hour=0, minute=0, second=0)
    end = end.replace(year=end.year + 1) - timedelta(seconds=1)
    return int(end.replace(tzinfo=timezone.utc).timestamp())

# Returns current timestamp
def now_ts() -> int:
    return int(datetime.now().replace(tzinfo=timezone.utc).timestamp())

# Returns timestamp of the last second of the day of given timestamp
def day_end(timestamp: int) -> int:
    end = datetime.utcfromtimestamp(timestamp).replace(hour=23, minute=59, second=59)
    return int(end.replace(tzinfo=timezone.utc).timestamp())
# -------------------------------------------------------------------------------------------------------------------


import logging
from datetime import datetime
from decimal import Decimal, InvalidOperation
from PySide6.QtCore import Qt, QDate
from jal.constants import BookAccount, MarketDataFeed, AssetData, PredefinedAsset
from jal.db.db import JalDB
from jal.db.helpers import format_decimal, year_begin, year_end
from jal.db.country import JalCountry
from jal.db.tag import JalTag
from jal.widgets.helpers import ts2d


# Helper function to convert db timestamp string into an integer and replace it as 0 if error happens
def db_timestamp2int(timestamp_string: str) -> int:
    try:
        timestamp = int(timestamp_string)
    except ValueError:
        timestamp = 0
    return timestamp


class JalAsset(JalDB):
    db_cache = []

    def __init__(self, asset_id: int = 0, data: dict = None, search: bool = False, create: bool = False) -> None:
        super().__init__(cached=True)
        if not JalAsset.db_cache:
            self._fetch_data()
        try:
            self._id = int(asset_id)
        except (TypeError, ValueError):
            self._id = 0
        if self._valid_data(data, search, create):
            if search:
                self._id = self._find_asset(data)
            if create and not self._id:   # If we haven't found peer before and requested to create new record
                query = self._exec(
                    "INSERT INTO assets (type_id, full_name, isin, country_id) "
                    "VALUES (:type, :full_name, :isin, coalesce((SELECT id FROM countries WHERE code=:country), 0))",
                    [(":type", data['type']), (":full_name", data['name']),
                     (":isin", data['isin']), (":country", data['country'])], commit=True)
                self._id = query.lastInsertId()
                self._fetch_data()
        self._data = next((x for x in self.db_cache if x['id'] == self._id), None)
        self._type = self._data['type_id'] if self._data is not None else None
        self._name = self._data['full_name'] if self._data is not None else ''
        self._isin = self._data['isin'] if self._data is not None else None
        self._country = JalCountry(self._data['country_id']) if self._data is not None else JalCountry(0)
        self._reg_number = self._data.get('data', {}).get(AssetData.RegistrationCode, '') if self._data is not None else ''
        self._expiry = int(self._data.get('data', {}).get(AssetData.ExpiryDate, 0)) if self._data is not None else 0
        self._principal = self._data.get('data', {}).get(AssetData.PrincipalValue, '') if self._data is not None else ''
        self._principal = Decimal(self._principal) if self._principal else Decimal('0')
        try:
            self._tag = JalTag(int(self._data.get('data', {}).get(AssetData.Tag, 0)))
        except (AttributeError, ValueError, TypeError):
            self._tag = JalTag(0)

    def invalidate_cache(self):
        self._fetch_data()

    # JalAsset maintains single cache available for all instances
    @classmethod
    def class_cache(cls) -> True:
        return True

    def _fetch_data(self):
        JalAsset.db_cache = []
        query = self._exec("SELECT * FROM assets ORDER BY id")
        while query.next():
            asset_data = self._read_record(query, named=True)
            asset_data['symbols'] = []
            symbols_query = self._exec("SELECT * FROM asset_tickers WHERE asset_id=:id", [(":id", asset_data['id'])])
            while symbols_query.next():
                symbol = self._read_record(symbols_query, named=True)
                del symbol['id']
                del symbol['asset_id']
                asset_data['symbols'].append(symbol)
            extra_data = {}
            data_query = self._exec("SELECT datatype, value FROM asset_data WHERE asset_id=:id ORDER BY datatype",
                                    [(":id", asset_data['id'])])
            while data_query.next():
                datatype, value = self._read_record(data_query)
                extra_data[datatype] = value
            if extra_data:
                asset_data['data'] = extra_data
            JalAsset.db_cache.append(asset_data)

    def dump(self) -> dict:
        return self._data

    def id(self) -> int:
        return self._id

    def type(self) -> int:
        return self._type

    def name(self) -> str:
        return self._name

    def isin(self) -> str:
        return self._isin

    # Returns asset symbol for given currency or all symbols if no currency is given
    def symbol(self, currency: int = None) -> str:
        if self._data is None:
            return ''
        currency = None if self._type == PredefinedAsset.Money else currency  # Money have one unique symbol
        if currency is None:
            return ','.join([x['symbol'] for x in self._data['symbols'] if x['active'] == 1])  # concatenate all symbols via comma
        else:
            symbol = [x['symbol'] for x in self._data['symbols'] if x['active'] == 1 and x['currency_id'] == currency]
            return ''.join(x for x in symbol)   # return symbol or empty string (there shouldn't be more than one)

    def add_symbol(self, symbol: str, currency_id: int=None, note: str='', data_source: int=MarketDataFeed.NA) -> None:
        existing = self._read("SELECT id, symbol, description, quote_source FROM asset_tickers "
                              "WHERE asset_id=:asset_id AND symbol=:symbol AND currency_id IS :currency",
                              [(":asset_id", self._id), (":symbol", symbol), (":currency", currency_id)], named=True)
        if existing is None:  # Deactivate old symbols and create a new one
            _ = self._exec("UPDATE asset_tickers SET active=0 WHERE asset_id=:asset_id AND currency_id IS :currency",
                           [(":asset_id", self._id), (":currency", currency_id)])
            _ = self._exec(
                "INSERT INTO asset_tickers (asset_id, symbol, currency_id, description, quote_source) "
                "VALUES (:asset_id, :symbol, :currency, :note, :data_source)",
                [(":asset_id", self._id), (":symbol", symbol), (":currency", currency_id),
                 (":note", note), (":data_source", data_source)])
        else:  # Update data for existing symbol
            if not existing['description']:
                _ = self._exec("UPDATE asset_tickers SET description=:note WHERE id=:id",
                               [(":note", note), (":id", existing['id'])])
            if existing['quote_source'] == MarketDataFeed.NA:
                _ = self._exec("UPDATE asset_tickers SET quote_source=:data_source WHERE id=:id",
                               [(":data_source", data_source), (":id", existing['id'])])
        self._fetch_data()

    # Returns country object for the asset
    def country(self) -> JalCountry:
        return self._country

    def country_name(self) -> str:
        return self._country.name()

    # Returns tuple in form of (timestamp:int, quote:Decimal) that contains last found quotation in given currency.
    # Returns (timestamp, 1) if quotation is requested relative to itself
    # Returned timestamp might be less than given. Returns (0, 0) if no quotation information present in db.
    def quote(self, timestamp: int, currency_id: int) -> tuple:
        if self._id == currency_id:
            return timestamp, Decimal('1')
        quote = self._read("SELECT timestamp, quote FROM quotes WHERE asset_id=:asset_id "
                           "AND currency_id=:currency_id AND timestamp<=:timestamp ORDER BY timestamp DESC LIMIT 1",
                           [(":asset_id", self._id), (":currency_id", currency_id), (":timestamp", timestamp)])
        if quote is None:
            if self._type == PredefinedAsset.Money and currency_id != self.get_base_currency(timestamp):  # find a cross-rate
                rate1 = self.quote(timestamp, self.get_base_currency(timestamp))[1]
                rate2 = JalAsset(currency_id).quote(timestamp, self.get_base_currency(timestamp))[1]
                rate = 0 if rate2 == Decimal('0') else rate1 / rate2
                return timestamp, rate
            else:
                logging.warning(self.tr("There are no quote/rate for ") +
                                f"{self.symbol(currency_id)} ({JalAsset(currency_id).symbol()}) {ts2d(timestamp)}")
                return 0, Decimal('0')
        return int(quote[0]), Decimal(quote[1])

    # Return a list of tuples (timestamp:int, quote:Decimal) of all quotes available for asset
    # for time interval begin-end
    def quotes(self, begin: int, end: int, currency_id: int) -> list:
        quotes = []
        query = self._exec(
            "SELECT timestamp, quote FROM quotes WHERE asset_id=:asset_id "
            "AND currency_id=:currency_id AND timestamp>=:begin AND timestamp<=:end ORDER BY timestamp",
            [(":asset_id", self._id), (":currency_id", currency_id), (":begin", begin), (":end", end)])
        while query.next():
            timestamp, quote = self._read_record(query, cast=[int, Decimal])
            quotes.append((timestamp, quote))
        return quotes

    # Returns tuple (begin_timestamp: int, end_timestamp: int) that defines timestamp range for which quotest are
    # available in database for given currency
    def quotes_range(self, currency_id: int) -> tuple:
        try:
            begin, end = self._read("SELECT MIN(timestamp), MAX(timestamp) FROM quotes "
                                    "WHERE asset_id=:asset_id AND currency_id=:currency_id",
                                    [(":asset_id", self._id), (":currency_id", currency_id)])
        except TypeError:
            begin = end = 0
        begin = db_timestamp2int(begin)
        end = db_timestamp2int(end)
        return begin, end

    # Returns a quote source id defined for given currency (currency_id can be None)
    def quote_source(self, currency_id: int) -> int:
        source_id = self._read("SELECT quote_source FROM asset_tickers "
                               "WHERE asset_id=:asset AND currency_id IS :currency",
                               [(":asset", self._id), (":currency", currency_id)])
        if source_id is None:
            return MarketDataFeed.NA
        else:
            return source_id

    # Returns a dict (ID-Name) of all available data sources
    @classmethod
    def get_sources_list(cls) -> dict:
        return MarketDataFeed().get_all_names()

    # Set quotations for given currency_id. Quotations is a list of {'timestamp':int, 'quote':Decimal} values
    def set_quotes(self, quotations: list, currency_id: int) -> None:
        data = [x for x in quotations if x['timestamp'] is not None and x['quote'] is not None]  # Drop Nones
        if data:
            for quote in quotations:
                _ = self._exec("INSERT OR REPLACE INTO quotes (asset_id, currency_id, timestamp, quote) "
                               "VALUES(:asset_id, :currency_id, :timestamp, :quote)",
                               [(":asset_id", self._id), (":currency_id", currency_id),
                                (":timestamp", quote['timestamp']), (":quote", format_decimal(quote['quote']))])
            begin = min(data, key=lambda x: x['timestamp'])['timestamp']
            end = max(data, key=lambda x: x['timestamp'])['timestamp']
            self.commit()
            logging.info(self.tr("Quotations were updated: ") +
                         f"{self.symbol(currency_id)} ({JalAsset(currency_id).symbol()}) {ts2d(begin)} - {ts2d(end)}")

    # returns expiration timestamp
    def expiry(self) -> int:
        return self._expiry

    # Returns number of days before expiration (negative value if asset is expired)
    def days2expiration(self) -> int:
        if self._expiry == 0:
            return 0
        expiry_date = datetime.utcfromtimestamp(self._expiry)
        days_remaining = int((expiry_date - datetime.utcnow()).total_seconds() / 86400)
        return days_remaining

    def reg_number(self):
        return self._reg_number

    def principal(self) -> Decimal:
        return self._principal

    def tag(self) -> JalTag:
        return self._tag

    def set_tag(self, tag_id: int) -> None:
        if self._tag.id() == tag_id:
            return
        _ = self._exec("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                       "VALUES(:asset_id, :datatype, :expiry)",
                       [(":asset_id", self._id), (":datatype", AssetData.Tag), (":expiry", str(tag_id))])
        self._tag = JalTag(tag_id)
        self._fetch_data()

    # Updates relevant asset data fields with information provided in data dictionary
    def update_data(self, data: dict) -> None:
        updaters = {
            'isin': self._update_isin,
            'name': self._update_name,
            'country': self._update_country,
            'reg_number': self._update_reg_number,
            'expiry': self._update_expiration,
            'principal': self._update_principal
        }
        if not self._id:
            return
        for key in data:
            if data[key]:
                try:
                    updaters[key](data[key])
                except KeyError:  # No updater for this key is present
                    continue
        self._fetch_data()

    def _update_isin(self, new_isin: str) -> None:
        if self._isin:
            if new_isin != self._isin:
                logging.error(self.tr("Unexpected attempt to update ISIN for ")
                              + f"{self.symbol()}: {self._isin} -> {new_isin}")
        else:
            _ = self._exec("UPDATE assets SET isin=:new_isin WHERE id=:id",
                           [(":new_isin", new_isin), (":id", self._id)])
            self._isin = new_isin

    def _update_name(self, new_name: str) -> None:
        if not self._name:
            _ = self._exec("UPDATE assets SET full_name=:new_name WHERE id=:id",
                           [(":new_name", new_name), (":id", self._id)])
            self._name = new_name

    def _update_country(self, new_code: str) -> None:
        if new_code.lower() != self._country.code().lower():
            new_country = JalCountry(data={'code': new_code.lower()}, search=True)
            if new_country.id():
                _ = self._exec("UPDATE assets SET country_id=:new_country_id WHERE id=:asset_id",
                               [(":new_country_id", new_country.id()), (":asset_id", self._id)])
                self._country_id = new_country.id()
                logging.info(self.tr("Country updated for ")
                             + f"{self.symbol()}: {self._country.name()} -> {new_country.name()}")

    def _update_reg_number(self, new_number: str) -> None:
        if new_number != self._reg_number:
            _ = self._exec("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                           "VALUES(:asset_id, :datatype, :reg_number)",
                           [(":asset_id", self._id), (":datatype", AssetData.RegistrationCode),
                            (":reg_number", new_number)])
            self._reg_number = new_number
            logging.info(self.tr("Reg.number updated for ")
                         + f"{self.symbol()}: {self._reg_number} -> {new_number}")

    def _update_expiration(self, new_expiration: int) -> None:
        _ = self._exec("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                       "VALUES(:asset_id, :datatype, :expiry)",
                       [(":asset_id", self._id), (":datatype", AssetData.ExpiryDate), (":expiry", str(new_expiration))])
        self._expiry = new_expiration

    def _update_principal(self, principal: str) -> None:
        if self._type != PredefinedAsset.Bond:
            return
        try:
            principal = Decimal(principal)
        except InvalidOperation:
            return
        if principal > Decimal('0'):
            _ = self._exec("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                           "VALUES(:asset_id, :datatype, :principal)",
                           [(":asset_id", self._id), (":datatype", AssetData.PrincipalValue),
                            (":principal", str(principal))])

    def _valid_data(self, data: dict, search: bool = False, create: bool = False) -> bool:
        if data is None:
            return False
        data['isin'] = data['isin'] if 'isin' in data else ''
        data['name'] = data['name'] if 'name' in data else ''
        data['country'] = data['country'] if 'country' in data else ''
        data['symbol'] = data['symbol'] if 'symbol' in data else ''
        data['reg_number'] = data['reg_number'] if 'reg_number' in data else ''
        return True

    def _find_asset(self, data: dict) -> int:
        id = None
        if data['isin']:
            # Select either by ISIN if no symbol given OR by both ISIN & symbol
            id = self._read("SELECT id FROM assets_ext "
                            "WHERE ((isin=:isin OR isin='') AND symbol=:symbol) OR (isin=:isin AND :symbol='')",
                            [(":isin", data['isin']), (":symbol", data['symbol'])])
            if id is None and data['symbol']:  # Make one more try by ISIN only if no match for ISIN+Symbol due to symbol change
                id = self._read("SELECT id FROM assets_ext WHERE isin=:isin", [(":isin", data['isin'])])
            if id is None:
                return 0
            else:
                return id
        if data['reg_number']:
            id = self._read("SELECT asset_id FROM asset_data WHERE datatype=:datatype AND value=:reg_number",
                            [(":datatype", AssetData.RegistrationCode), (":reg_number", data['reg_number'])])
            if id is not None:
                return id
        if data['symbol']:
            if 'type' in data:
                if 'expiry' in data:
                    id = self._read("SELECT a.id FROM assets_ext a "
                                    "LEFT JOIN asset_data d ON a.id=d.asset_id AND d.datatype=:datatype "
                                    "WHERE symbol=:symbol COLLATE NOCASE AND type_id=:type AND value=:value",
                                    [(":datatype", AssetData.ExpiryDate), (":symbol", data['symbol']),
                                     (":type", data['type']), (":value", data['expiry'])])
                else:
                    id = self._read("SELECT id FROM assets_ext "
                                    "WHERE symbol=:symbol COLLATE NOCASE and type_id=:type",
                                    [(":symbol", data['symbol']), (":type", data['type'])])
            else:
                id = self._read("SELECT id FROM assets_ext WHERE symbol=:symbol COLLATE NOCASE",
                                [(":symbol", data['symbol'])])
            if id is not None:
                return id
        if data['name']:
            id = self._read("SELECT id FROM assets_ext WHERE full_name=:name COLLATE NOCASE",
                            [(":name", data['name'])])
        if id is None:
            return 0
        else:
            return id

    # Method returns a list of {"asset": JalAsset, "currency" currency_id} that describes assets involved into ledger
    # operations between begin and end timestamps or that have non-zero value in ledger
    @classmethod
    def get_active_assets(cls, begin: int, end: int) -> list:
        assets = []
        query = cls._exec("SELECT MAX(l.id) AS id, l.asset_id, a.currency_id "
                          "FROM ledger l LEFT JOIN accounts a ON a.id=l.account_id "
                          "WHERE l.book_account=:assets "
                          "GROUP BY l.asset_id, a.currency_id "
                          "HAVING l.amount_acc!='0' OR (l.timestamp>=:begin AND l.timestamp<=:end)",
                          [(":assets", BookAccount.Assets), (":begin", begin), (":end", end)])
        while query.next():
            try:
                _id, asset_id, currency_id = super(JalAsset, JalAsset)._read_record(query, cast=[int, int, int])
            except TypeError:  # Skip if None is returned (i.e. there are no assets)
                continue
            assets.append({"asset": JalAsset(asset_id), "currency": currency_id})
        return assets

    # Method returns a list of JalAsset objects that describe all assets defined in ledger
    @classmethod
    def get_assets(cls) -> list:
        assets = []
        query = cls._exec("SELECT id FROM assets")
        while query.next():
            assets.append(JalAsset(super(JalAsset, JalAsset)._read_record(query, cast=[int])))
        return assets

    # Method returns a list of JalAsset objects that describe currencies defined in ledger
    @classmethod
    def get_currencies(cls) -> list:
        currencies = []
        query = cls._exec("SELECT id FROM currencies")
        while query.next():
            currencies.append(JalAsset(super(JalAsset, JalAsset)._read_record(query, cast=[int])))
        return currencies

    # Returns id of the base currency that was in effect for given timestamp or current base currency (for now) if
    # timestamp isn't given
    @classmethod
    def get_base_currency(cls, timestamp: int=None) -> int:
        if timestamp is None:
            timestamp = QDate.currentDate().startOfDay(Qt.UTC).toSecsSinceEpoch()
        base_id = cls._read("SELECT currency_id FROM base_currency WHERE since_timestamp<=:timestamp "
                            "ORDER BY since_timestamp DESC LIMIT 1", [(":timestamp", timestamp)])
        try:
            base_id = int(base_id)
        except TypeError:
            base_id = 0
        return base_id

    # Return a list of (timestamp, currency_id) tuples that represent currency valid currency IDs that were in force
    # after between beginning_of_the_year(begin) and end_of_the_year(end) timestamps.
    # Begin and end of year it required to cover full tax year.
    @classmethod
    def get_base_currency_history(cls, begin: int, end: int) -> list:
        history = [(year_begin(begin), cls.get_base_currency(year_begin(begin)))]
        query = cls._exec("SELECT since_timestamp, currency_id FROM base_currency "
                          "WHERE since_timestamp>:begin AND since_timestamp<=:end ORDER BY since_timestamp",
                          [(":begin", year_begin(begin)), (":end", year_end(end))])
        while query.next():
            history.append(cls._read_record(query, cast=[int, int]))
        return history



import logging
import re
from datetime import timezone
from jal.data_import.statement_xls import StatementXLS
from jal.data_import.statement import FOF, Statement_ImportError


JAL_STATEMENT_CLASS = "StatementVTB"

# ----------------------------------------------------------------------------------------------------------------------
class StatementVTB(StatementXLS):
    PeriodPattern = (7, 1, r"Отчет Банка ВТБ \(ПАО\) за период с (?P<S>\d\d\.\d\d\.\d\d\d\d) по (?P<E>\d\d\.\d\d\.\d\d\d\d) о сделках, .*")
    AccountPattern = (9, 7, None)
    HeaderCol = 1
    money_section = "Отчет об остатках денежных средств"
    money_columns = {
        "name": "Валюта",
        "cash_end": "Плановый",
    }
    asset_section = "Отчет об остатках ценных бумаг"
    asset_columns = {
        "name": "Наименование ценной бумаги, \n№ гос. регистрации, ISIN",
        "currency": "Валюта цены \n\(номинала для облигаций\)"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("VTB Investments")
        self.icon_name = "vtb.ico"
        self.filename_filter = self.tr("VTB statement (*.xls)")
        self.account_end_balance = {}

    def _load_currencies(self):
        cnt = 0
        row, headers = self.find_section_start(self.money_section, self.money_columns, header_height=3)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row].startswith('Сумма') or self._statement[self.HeaderCol][row] == '':
                break
            currency = self._statement[headers['name']][row]
            try:
                code = self.currency_substitutions[currency]
            except KeyError:
                code = currency
            self.currency_id(code)
            self.account_end_balance[code] = self._statement[headers['cash_end']][row]
            cnt += 1
            row += 1
        logging.info(self.tr("Account currencies loaded: ") + f"{cnt}")

    # Planned money amount already loaded in _load_currencies(). Here it is only put in account data
    def _load_money(self):
        for currency in self.account_end_balance:
            account_id = self._find_account_id(self._account_number, currency)
            account = [x for x in self._data[FOF.ACCOUNTS] if x['id'] == account_id][0]
            account["cash_end"] = self.account_end_balance[currency]

    def _load_assets(self):
        AssetPattern = r"^(?P<name>.*), (?P<reg_number>.*), (?P<isin>.*)$"
        cnt = 0
        row, headers = self.find_section_start(self.asset_section, self.asset_columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row].startswith('ИТОГО') or self._statement[self.HeaderCol][row] == '':
                break
            if self._statement[self.HeaderCol][row].startswith(('АКЦИЯ', 'ОБЛИГАЦИЯ', 'ПАЙ')):
                row += 1  # This is an asset class separator - move to the next row
                continue
            asset_name = self._statement[headers['name']][row]
            parts = re.match(AssetPattern, asset_name, re.IGNORECASE)
            if parts is None:
                raise Statement_ImportError(self.tr("Can't parse asset name ") + f"'{asset_name}'")
            asset_data = parts.groupdict()
            if len(asset_data) != AssetPattern.count("(?P<"):  # check that expected number of groups was matched
                raise Statement_ImportError(self.tr("Asset name miss some data ") + f"'{asset_name}'")
            currency = self._statement[headers['currency']][row]
            try:
                currency_code = self.currency_substitutions[currency]
            except KeyError:
                currency_code = currency
            asset_id = self.asset_id({'isin': asset_data['isin'], 'reg_number': asset_data['reg_number'],
                                      'currency': self.currency_id(currency_code), 'search_offline': True,
                                      'search_online': "MOEX"})
            asset = self._find_in_list(self._data[FOF.ASSETS], 'id', asset_id)
            asset['broker_name'] = asset_name
            cnt += 1
            row += 1
        logging.info(self.tr("Securities loaded: ") + f"{cnt}")

    def _strip_unused_data(self):
        for asset in self._data[FOF.ASSETS]:
            self.drop_extra_fields(asset, ['broker_name'])

    def _load_deals(self):
        cnt = 0
        columns = {
            "asset_name": "Наименование ценной бумаги, \n№ гос\. Регистрации, ISIN",
            "number": "№ сделки",
            "datetime": "Дата и время заключения сделки",
            "B/S": "Вид сделки",
            "price": "Цена\n\(% для облигаций\)",
            "currency": "Валюта цены\n \(номинала для облигаций\)",
            "qty": "Количество \n\(шт\)",
            "amount": "Сумма сделки в валюте расчетов\n \(с учетом НКД для облигаций\)  ",
            "accrued_int": "НКД\nпо сделке в валюте расчетов",
            "settlement": " Плановая дата поставки",
            "fee1": "Комиссия Банка за расчет по сделке",
            "fee2": "Комиссия Банка за заключение сделки"
        }
        row, headers = self.find_section_start("Заключенные в отчетном периоде сделки с ценными бумагами", columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            asset_name = self._statement[headers['asset_name']][row]
            assets = [x for x in self._data[FOF.ASSETS] if x.get('broker_name') == asset_name]
            if len(assets) != 1:
                raise Statement_ImportError(self.tr("No asset match in deals for ") + f"'{asset_name}'")
            asset_id = assets[0]['id']
            if self._statement[headers['B/S']][row] == 'Покупка':
                qty = self._statement[headers['qty']][row]
                bond_interest = -self._statement[headers['accrued_int']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
                bond_interest = self._statement[headers['accrued_int']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            deal_number = self._statement[headers['number']][row]
            price = self._statement[headers['price']][row]
            currency = self._statement[headers['currency']][row]
            fee = self._statement[headers['fee1']][row] + self._statement[headers['fee2']][row]
            amount = self._statement[headers['amount']][row]
            if abs(abs(price * qty) - amount) >= self.RU_PRICE_TOLERANCE:
                price = abs(amount / qty)
            timestamp = int(self._statement[headers['datetime']][row].replace(tzinfo=timezone.utc).timestamp())
            settlement = int(self._statement[headers['settlement']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, currency)
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            if bond_interest != 0:
                new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
                payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                           "number": deal_number, "asset": asset_id, "amount": bond_interest, "description": "НКД"}
                self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Trades loaded: ") + f"{cnt}")

    def _load_cash_transactions(self):
        cnt = 0
        columns = {
            "date": "Дата",
            "type": "Тип операции",
            "amount": "Сумма",
            "currency": "Валюта",
            "description": "Комментарий"
        }
        operations = {
            'Зачисление денежных средств': self.transfer_in,
            'Списание денежных средств': self.transfer_out,
            'Купонный доход': self.interest,
            'Сальдо расчётов по сделкам с ценными бумагами': None,  # These operations are results of trades
            'Вознаграждение Брокера': None
        }
        row, headers = self.find_section_start("Движение денежных средств", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':  # Stop if there are no next date available
                break
            operation = self._statement[headers['type']][row]
            if operation == '':  # Skip market header as of now
                row += 1
                continue
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported cash transaction ") + f"'{operation}'")
            timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            amount = self._statement[headers['amount']][row]
            description = self._statement[headers['description']][row]
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            if operations[operation] is not None:
                operations[operation](timestamp, account_id, amount, description)
            cnt += 1
            row += 1
        logging.info(self.tr("Cash operations loaded: ") + f"{cnt}")

    def transfer_in(self, timestamp, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def interest(self, timestamp, account_id, amount, description):
        BondInterestPattern = r"^Куп\. дох\. по обл\. .* (?P<reg_number>\S*), \S*\..*$"
        parts = re.match(BondInterestPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse bond interest description ") + f"'{description}'")
        interest_data = parts.groupdict()
        asset_id = self._find_in_list(self._data[FOF.ASSETS_DATA], 'reg_number', interest_data['reg_number'])['asset']
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                   "asset": asset_id, "amount": amount, "description": description}
        self._data[FOF.ASSET_PAYMENTS].append(payment)



from __future__ import annotations
from datetime import datetime
from decimal import Decimal

from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QHeaderView
from jal.db.helpers import now_ts, day_end
from jal.db.tree_model import AbstractTreeItem, ReportTreeModel
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.operations import LedgerTransaction, Transfer, CorporateAction
from jal.widgets.delegates import GridLinesDelegate, FloatDelegate, TimestampDelegate
from jal.widgets.helpers import ts2d


# ----------------------------------------------------------------------------------------------------------------------
OLD_QUOTE_AGE = 7   # when mark line as italic for old quote
# ----------------------------------------------------------------------------------------------------------------------
class AssetTreeItem(AbstractTreeItem):
    def __init__(self, data=None, parent=None, group=''):
        super().__init__(parent, group)
        if data is None:
            self._data = {
                'header': '', 'currency_id': 0, 'currency': '', 'account_id': 0, 'account': '', 'asset_id': 0,
                'tag': '', 'asset_is_currency': False, 'asset': '', 'asset_name': '', 'country_id': 0, 'country': '',
                'since': 0, 'qty': None, 'value_i': Decimal('0'), 'paid': Decimal('0'),
                'open_quote': None, 'quote': None, 'quote_ts': Decimal('0'), 'quote_a': Decimal('0'),
                'value': Decimal('0'), 'value_common': Decimal('0'), 'p/l': Decimal('0'), 'p/l%': Decimal('0'),
                'font': 'bold', 'quote_age': 0, 'share': Decimal('0')
            }
        else:
            self._data = data.copy()
            self._data['value'] = self._data['quote'] * self._data['qty']
            self._data['value_common'] = self._data['quote_a'] * self._data['qty'] if self._data['quote_a'] else Decimal('0')
            self._data['p/l'] = self._data['quote'] * self._data['qty'] - self._data['value_i'] if not self._data['asset_is_currency'] else Decimal('0')
            self._data['p/l%'] = Decimal('100') * (self._data['quote'] * self._data['qty'] / self._data['value_i'] - 1) if self._data['value_i'] != Decimal('0') else Decimal('0')

    def _calculateGroupTotals(self, child_data):
        self._data['header'] = child_data.get(self._group.strip("_id"), '<none>')
        self._data['currency'] = child_data['currency']
        self._data['account'] = child_data['account']
        self._data['asset'] = child_data['asset']
        self._data['country'] = child_data['country']
        self._data['tag'] = child_data['tag']
        self._data['paid'] += child_data['paid']
        self._data['value'] += child_data['value']
        self._data['value_common'] += child_data['value_common']
        self._data['p/l'] += child_data['p/l']
        self._data['p/l%'] = Decimal('100') * self._data['p/l'] / (self._data['value'] - self._data['p/l']) if self._data['value'] != self._data['p/l'] else Decimal('0')

    def _afterParentGroupUpdate(self, group_data):
        self._data['share'] = Decimal('100') * self._data['value_common'] / group_data['value_common'] if group_data['value_common'] else Decimal('0')

    def details(self):
        return self._data

    # assigns group value if this tree item is a group item
    def setGroupValue(self, value):
        if self._group:
            self._data[self._group] = value

    def getGroup(self):
        if self._group:
            return self._group, self._data[self._group]
        else:
            return None

    # returns an element of tree that will provide right group parent for 'item' with given 'group_fields'
    def getGroupLeaf(self, group_fields: list, item: AssetTreeItem) -> AssetTreeItem:
        if group_fields:
            group_item = None
            group_name = group_fields[0]
            for child in self._children:
                if child.details()[group_name] == item.details()[group_name]:
                    group_item = child
            if group_item is None:
                group_item = AssetTreeItem(None, parent=self, group=group_name)
                group_item.setGroupValue(item.details()[group_name])
                self._children.append(group_item)
            return group_item.getGroupLeaf(group_fields[1:], item)
        else:
            return self

# ----------------------------------------------------------------------------------------------------------------------
class HoldingsModel(ReportTreeModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._grid_delegate = None
        self._float_delegate = None
        self._float2_delegate = None
        self._float4_delegate = None
        self._profit_delegate = None
        self._date_delegate = None
        bold_font = QFont()
        bold_font.setBold(True)
        strikeout_font = QFont()
        strikeout_font.setStrikeOut(True)
        italic_font = QFont()
        italic_font.setItalic(True)
        self._fonts = {'normal': None, 'bold': bold_font, 'italic': italic_font, 'strikeout': strikeout_font}
        self._currency = 0
        self._only_active_accounts = True
        self._currency_name = ''
        self._date = day_end(now_ts())
        self._columns = [{'name': self.tr("Currency/Account/Asset"), 'field': 'header'},
                         {'name': self.tr("Asset Name"), 'field': 'asset_name'},
                         {'name': self.tr("Qty"), 'field': 'qty'},
                         {'name': self.tr("Since"), 'field': 'since'},
                         {'name': self.tr("Open"), 'field': 'open_quote'},
                         {'name': self.tr("Last"), 'field': 'quote'},
                         {'name': self.tr("Share, %"), 'field': 'share'},
                         {'name': self.tr("P/L, %"), 'field': 'p/l%'},
                         {'name': self.tr("P/L"), 'field': 'p/l'},
                         {'name': self.tr("Paid"), 'field': 'paid'},
                         {'name': self.tr("Value"), 'field': 'value'},
                         {'name': self.tr("Value, "), 'field': 'value_common'}]

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        value = super().headerData(section, orientation, role)
        if orientation == Qt.Horizontal and role == Qt.DisplayRole and section == self.fieldIndex('value_common'):
            value += self._currency_name
        return value

    def data(self, index, role=Qt.DisplayRole):
        assert index.column() in range(len(self._columns))
        try:
            if not index.isValid():
                return None
            item = index.internalPointer()
            if role == Qt.DisplayRole:
                return item.details().get(self._columns[index.column()]['field'], None)
            if role == Qt.FontRole:
                return self._fonts.get(item.details()['font'], None)
            if role == Qt.ToolTipRole:
                return self.data_tooltip(item.details())
            return None
        except Exception as e:
            print(e)

    def data_tooltip(self, data):
        if data['quote_age'] > OLD_QUOTE_AGE:
            return self.tr("Last quote date: ") + ts2d(int(data['quote_ts']))
        return None

    def configureView(self):
        for field in [x['field'] for x in self._columns]:
            if field == 'asset_name':
                self._view.header().setSectionResizeMode(self.fieldIndex(field), QHeaderView.Stretch)
            else:
                self._view.header().setSectionResizeMode(self.fieldIndex(field), QHeaderView.ResizeToContents)
        self._grid_delegate = GridLinesDelegate(self._view)
        self._date_delegate = TimestampDelegate(display_format='%d/%m/%Y', parent=self._view)
        self._float_delegate = FloatDelegate(0, allow_tail=True, parent=self._view)
        self._float2_delegate = FloatDelegate(2, allow_tail=False, parent=self._view)
        self._float4_delegate = FloatDelegate(4, allow_tail=False, parent=self._view)
        self._profit_delegate = FloatDelegate(2, allow_tail=False, colors=True, parent=self._view)
        self._view.setItemDelegateForColumn(self.fieldIndex('header'), self._grid_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('asset_name'), self._grid_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('qty'), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('since'), self._date_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('open_quote'), self._float4_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('quote'), self._float4_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('share'), self._float2_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('p/l%'), self._profit_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('p/l'), self._profit_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('paid'), self._float2_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('value'), self._float2_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex('value_common'), self._float2_delegate)
        super().configureView()

    def updateView(self, currency_id, date, grouping, show_inactive):
        update = False
        if self._currency != currency_id:
            self._currency = currency_id
            self._currency_name = JalAsset(currency_id).symbol()
            update = True
        if self._date != date.endOfDay(Qt.UTC).toSecsSinceEpoch():
            self._date = date.endOfDay(Qt.UTC).toSecsSinceEpoch()
            update = True
        if self._only_active_accounts == show_inactive:  # The logic is reversed inside the report
            self._only_active_accounts = not self._only_active_accounts
            update = True
        if self.setGrouping(grouping) or update:
            self.prepareData()

    def get_data_for_tax(self, index):
        if not index.isValid():
            return None
        data = index.internalPointer().details()
        return data['account_id'], data['asset_id'], data['currency_id'], data['qty']

    # Returns tuple of two values for open position in 'asset' for 'account' on given date 'end_ts':
    # 1st (int) - the earliest timestamp of open position
    # 2nd (Decimal) = 0 (not implemented - the amount of payments that were accumulated for given asset)
    def get_asset_history_payments(self, account: JalAccount, asset: JalAsset, end_ts: int) -> (int, Decimal):
        trades = account.open_trades_list(asset, end_ts)
        since = min([x['operation'].timestamp() for x in trades])
        amount = account.asset_payments_amount(asset, since, end_ts)
        for trade in trades:
            operation = trade['operation']
            if operation.type() == LedgerTransaction.Transfer:
                transfer_out = LedgerTransaction().get_operation(operation.type(), operation.id(), Transfer.Outgoing)
                since_new, amount_new = self.get_asset_history_payments(transfer_out.account(), asset, transfer_out.timestamp()-1)  # get position just before the transfer
            elif operation.type() == LedgerTransaction.CorporateAction and operation.subtype() == CorporateAction.Split:
                since_new, amount_new = self.get_asset_history_payments(account, asset, operation.timestamp()-1)  # get position just before the split
            else:
                continue
            since = min(since, since_new)
            amount += amount_new
        return since, amount

    # Populate table 'holdings' with data calculated for given parameters of model: _currency, _date,
    def prepareData(self):
        holdings = []
        accounts = JalAccount.get_all_accounts(investing_only=True, active_only=self._only_active_accounts)
        for account in accounts:
            account_holdings = []
            assets = account.assets_list(self._date)
            rate = JalAsset(account.currency()).quote(self._date, self._currency)[1]
            all_fields = ['currency', 'account', 'asset']
            display_fields = [y for y in all_fields if y not in [x.strip("_id") for x in self._groups]]
            for asset_data in assets:
                asset = asset_data['asset']
                quote_ts, quote = asset.quote(self._date, account.currency())
                quote_age = int((datetime.utcnow() - datetime.utcfromtimestamp(quote_ts)).total_seconds() / 86400)
                since, payments_amount = self.get_asset_history_payments(account, asset, self._date)
                font = 'italic' if quote_age > OLD_QUOTE_AGE else 'normal'
                font = 'strikeout' if asset.days2expiration() < 0 else font
                expiry_header = self.tr("Exp:")
                expiry_text = f" [{expiry_header} {ts2d(asset.expiry())}]" if asset.expiry() else ''
                record = {
                    "currency_id": account.currency(),                      # Fields 'x' and 'x_id' are used together for grouping
                    "currency": JalAsset(account.currency()).symbol(),
                    "account_id": account.id(),
                    "account": account.name(),
                    "asset_id": asset.id(),
                    "asset": asset.symbol(currency=account.currency()),
                    "asset_name": asset.name() + expiry_text,
                    "asset_is_currency": False,
                    "country_id": asset.country().id(),
                    "country": asset.country().name(),
                    "tag": asset.tag().name() if asset.tag().name() else self.tr("N/A"),
                    "since": since,
                    "qty": asset_data['amount'],
                    "value_i": asset_data['value'],
                    "paid": payments_amount,
                    "open_quote": asset_data['value'] / asset_data['amount'],
                    "quote": quote,
                    "quote_ts": quote_ts,
                    "quote_a": rate * quote,
                    "quote_age": quote_age,
                    "share": Decimal('0'),
                    "font": font
                }
                record['header'] = ': '.join([record[x] for x in display_fields])
                account_holdings.append(record)
            money = account.get_asset_amount(self._date, account.currency())
            if money:
                record = {
                    "currency_id": account.currency(),
                    "currency": JalAsset(account.currency()).symbol(),
                    "account_id": account.id(),
                    "account": account.name(),
                    "asset_id": account.currency(),
                    "asset": JalAsset(account.currency()).symbol(),
                    "asset_name": JalAsset(account.currency()).name(),
                    "asset_is_currency": True,
                    "country_id": JalAsset(account.currency()).country().id(),
                    "country": JalAsset(account.currency()).country().name(),
                    "tag": self.tr("Money"),
                    "since": 0,
                    "qty": money,
                    "value_i": Decimal('0'),
                    "paid": Decimal('0'),
                    "open_quote": None,
                    "quote": Decimal('1'),
                    "quote_ts": day_end(now_ts()),
                    "quote_a": rate,
                    "quote_age": 0,
                    "share": Decimal('0'),
                    "font": 'normal'
                }
                record['header'] = ': '.join([record[x] for x in display_fields])
                account_holdings.append(record)
            holdings += account_holdings
        # Sort data with respect to groups before building the tree
        sort_names = [x.removesuffix("_id") for x in self._groups]
        if 'asset' in sort_names:
            sort_names.insert(sort_names.index('asset'), 'asset_is_currency')  # Need to put currency at the end
        else:
            sort_names += ['asset_is_currency', 'asset']   # Sort by asset name for any kind of grouping
        holdings = sorted(holdings, key=lambda x: tuple([x[key_name] for key_name in sort_names]))
        # Add data records to the report tree
        self._root = AssetTreeItem()
        for position in holdings:
            new_item = AssetTreeItem(position)
            leaf = self._root.getGroupLeaf(self._groups, new_item)
            leaf.appendChild(new_item)
        super().prepareData()
